__author__ = "Simon Nilsson"

import multiprocessing

import numpy as np
from numba import (boolean, float32, float64, int64, jit, njit, prange, typed,
                   types)
from numba.typed import Dict, List
from numpy.lib.stride_tricks import as_strided
from statsmodels.tsa.stattools import adfuller, kpss, zivot_andrews

from simba.utils.errors import InvalidInputError

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from typing import get_type_hints

from simba.utils.read_write import find_core_cnt


class TimeseriesFeatureMixin(object):
    """
    Time-series methods focused on signal complexity in sliding windows. Mainly in time-domain - fft methods (through e.g. scipy)
    I've found so far has not been fast enough for rolling windows in large datasets.

    .. image:: _static/img/ts_runtimes.png
       :width: 1200
       :align: center

    .. note::
       Many method has numba typed `signatures <https://numba.pydata.org/numba-doc/latest/reference/types.html>`_ to decrease
       compilation time. Make sure to pass the correct dtypes as indicated by signature decorators.

    .. important::
       See references for mature packages computing more extensive timeseries measurements

       .. [1] `cesium <https://github.com/cesium-ml/cesium>`_.
       .. [2] `eeglib <https://github.com/Xiul109/eeglib>`_.
       .. [3] `antropy <https://github.com/raphaelvallat/antropy>`_.
       .. [4] `tsfresh <https://tsfresh.readthedocs.io>`_.
       .. [5] `pycaret <https://github.com/pycaret/pycaret>`_.
    """

    def __init__(self):
        pass

    @staticmethod
    @njit("(float32[:],)")
    def hjort_parameters(data: np.ndarray) -> (float, float, float):
        """
        Jitted compute of Hjorth parameters for a given time series data. Hjorth parameters describe
        mobility, complexity, and activity of a time series.

        :param numpy.ndarray data: A 1-dimensional numpy array containing the time series data.
        :return: tuple
            A tuple containing the following Hjorth parameters:
            - activity (float): The activity of the time series, which is the variance of the input data.
            - mobility (float): The mobility of the time series, calculated as the square root of the variance
              of the first derivative of the input data divided by the variance of the input data.
            - complexity (float): The complexity of the time series, calculated as the square root of the variance
              of the second derivative of the input data divided by the variance of the first derivative, and then
              divided by the mobility.

        :example:
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
        >>> TimeseriesFeatureMixin().hjort_parameters(data)
        >>> (2.5, 0.5, 0.4082482904638631)

        :math:`mobility = \sqrt{\\frac{dx_{var}}{x_{var}}}`

        :math:`complexity = \sqrt{\\frac{ddx_{var}}{dx_{var}} / mobility}`
        """

        def diff(x):
            return x[1:] - x[:-1]

        dx = np.diff(np.ascontiguousarray(data))
        ddx = np.diff(np.ascontiguousarray(dx))
        x_var, dx_var = np.var(data), np.var(dx)
        if (x_var <= 0) or (dx_var <= 0):
            return 0, 0, 0
        ddx_var = np.var(ddx)
        mobility = np.sqrt(dx_var / x_var)
        complexity = np.sqrt(ddx_var / dx_var) / mobility
        activity = np.var(data)

        return activity, mobility, complexity

    @staticmethod
    @njit("(float32[:], float64[:], int64)")
    def sliding_hjort_parameters(
        data: np.ndarray, window_sizes: np.ndarray, sample_rate: int
    ) -> np.ndarray:
        """
        Jitted compute of Hjorth parameters, including mobility, complexity, and activity, for
        sliding windows of varying sizes applied to the input data array.

        :param np.ndarray data: Input data array.
        :param np.ndarray window_sizes: Array of window sizes (in seconds).
        :param int sample_rate: Sampling rate of the data in samples per second.
        :return np.ndarray: An array containing Hjorth parameters for each window size and data point.
                   The shape of the result array is (3, data.shape[0], window_sizes.shape[0]).
                   The three parameters are stored in the first dimension (0 - mobility, 1 - complexity,
                   2 - activity), and the remaining dimensions correspond to data points and window sizes.

        """
        results = np.full((3, data.shape[0], window_sizes.shape[0]), -1.0)
        for i in range(window_sizes.shape[0]):
            window_size = int(window_sizes[i] * sample_rate)
            for l, r in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                sample = data[l:r]
                dx = np.diff(np.ascontiguousarray(sample))
                ddx = np.diff(np.ascontiguousarray(dx))
                x_var, dx_var = np.var(sample), np.var(dx)
                if (x_var <= 0) or (dx_var <= 0):
                    results[0, r + 1, i] = 0
                    results[1, r + 1, i] = 0
                    results[2, r + 1, i] = 0
                else:
                    ddx_var = np.var(ddx)
                    mobility = np.sqrt(dx_var / x_var)
                    complexity = np.sqrt(ddx_var / dx_var) / mobility
                    activity = np.var(sample)
                    results[0, r + 1, i] = mobility
                    results[1, r + 1, i] = complexity
                    results[2, r + 1, i] = activity

        return results.astype(np.float32)

    @staticmethod
    @njit([(float32[:], boolean), (float32[:], types.misc.Omitted(True))])
    def local_maxima_minima(data: np.ndarray, maxima: bool = True) -> np.ndarray:
        """
        Jitted compute of the local maxima or minima defined as values which are higher or lower than immediately preceding and proceeding time-series neighbors, repectively.
        Returns 2D np.ndarray with columns representing idx and values of local maxima.

        :param np.ndarray data: Time-series data.
        :param bool maxima: If True, returns maxima. Else, minima.
        :return np.ndarray: 2D np.ndarray with columns representing idx in input data in first column and values of local maxima in second column

        :example:
        >>> data = np.array([3.9, 7.5,  4.2, 6.2, 7.5, 3.9, 6.2, 6.5, 7.2, 9.5]).astype(np.float32)
        >>> TimeseriesFeatureMixin().local_maxima_minima(data=data, maxima=True)
        >>> [[1, 7.5], [4, 7.5], [9, 9.5]]
        >>> TimeseriesFeatureMixin().local_maxima_minima(data=data, maxima=False)
        >>> [[0, 3.9], [2, 4.2], [5, 3.9]]

        .. image:: _static/img/local_maxima_minima.png
           :width: 600
           :align: center

        """
        if not maxima:
            data = -data
        results = np.full((data.shape[0], 2), -1.0)
        if data[0] >= data[1]:
            if not maxima:
                results[0, :] = np.array([0, -data[0]])
            else:
                results[0, :] = np.array([0, data[0]])
        if data[-1] >= data[-2]:
            if not maxima:
                results[-1, :] = np.array([data.shape[0] - 1, -data[-1]])
            else:
                results[-1, :] = np.array([data.shape[0] - 1, data[-1]])
        for i in prange(1, data.shape[0] - 1):
            if data[i - 1] < data[i] > data[i + 1]:
                if not maxima:
                    results[i, :] = np.array([i, -data[i]])
                else:
                    results[i, :] = np.array([i, data[i]])

        return results[np.argwhere(results[:, 0].T != -1).flatten()]

    @staticmethod
    @njit("(float32[:], float64)")
    def crossings(data: np.ndarray, val: float) -> int:
        """
        Jitted compute of the count in time-series where sequential values crosses a defined value.


        .. image:: _static/img/crossings.png
           :width: 600
           :align: center

        :param np.ndarray data: Time-series data.
        :param float val: Cross value. E.g., to count the number of zero-crossings, pass `0`.
        :return int: Count of events where sequential values crosses ``val``.

        :example:
        >>> data = np.array([3.9, 7.5,  4.2, 6.2, 7.5, 3.9, 6.2, 6.5, 7.2, 9.5]).astype(np.float32)
        >>> TimeseriesFeatureMixin().crossings(data=data, val=7)
        >>> 5
        """

        cnt, last_val = 0, -1
        if data[0] > val:
            last_val = 1
        for i in prange(1, data.shape[0]):
            current_val = -1
            if data[i] > val:
                current_val = 1
            if last_val != current_val:
                cnt += 1
            last_val = current_val

        return cnt

    @staticmethod
    @njit("(float32[:], float64,  float64[:], int64,)")
    def sliding_crossings(
        data: np.ndarray, val: float, window_sizes: np.ndarray, sample_rate: int
    ) -> np.ndarray:
        """
        Compute the number of crossings over sliding windows in a data array.

        Computes the number of times a value in the data array crosses a given threshold
        value within sliding windows of varying sizes. The number of crossings is computed for each
        window size and stored in the result array where columns represents time windows.

        .. image:: _static/img/sliding_crossings.png
           :width: 1500
           :align: center

        :param np.ndarray data: Input data array.
        :param float val: Threshold value for crossings.
        :param np.ndarray window_sizes: Array of window sizes (in seconds).
        :param int sample_rate: Sampling rate of the data in samples per second.
        :return np.ndarray: An array containing the number of crossings for each window size and data point. The shape of the result array is (data.shape[0], window_sizes.shape[0]).
        """

        results = np.full((data.shape[0], window_sizes.shape[0]), -1.0)
        for i in prange(window_sizes.shape[0]):
            window_size = int(window_sizes[i] * sample_rate)
            for l, r in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                sample = data[l:r]
                cnt, last_val = 0, -1
                if sample[0] > val:
                    last_val = 1
                for j in prange(1, sample.shape[0]):
                    current_val = -1
                    if sample[j] > val:
                        current_val = 1
                    if last_val != current_val:
                        cnt += 1
                    last_val = current_val
                results[r - 1, i] = cnt

        return results.astype(np.int32)

    @staticmethod
    @njit("(float32[:], int64, int64, )", cache=True, fastmath=True)
    def percentile_difference(
        data: np.ndarray, upper_pct: int, lower_pct: int
    ) -> float:
        """
        Jitted compute of the difference between the ``upper`` and ``lower`` percentiles of the data as
        a percentage of the median value.

        .. note::
           Adapted from `cesium <https://github.com/cesium-ml/cesium>`_.


        .. image:: _static/img/percentile_difference.png
           :width: 600
           :align: center

        :parameter np.ndarray data: 1D array of representing time-series.
        :parameter int upper_pct: Upper-boundary percentile.
        :parameter int lower_pct: Lower-boundary percentile.
        :returns float: The difference between the ``upper`` and ``lower`` percentiles of the data as a percentage of the median value.

        :examples:
        >>> data = np.array([3.9, 7.5,  4.2, 6.2, 7.5, 3.9, 6.2, 6.5, 7.2, 9.5]).astype(np.float32)
        >>> TimeseriesFeatureMixin().percentile_difference(data=data, upper_pct=95, lower_pct=5)
        >>> 0.7401574764125177

        """

        upper_val, lower_val = np.percentile(data, upper_pct), np.percentile(
            data, lower_pct
        )
        return np.abs(upper_val - lower_val) / np.median(data)

    @staticmethod
    @njit("(float32[:], int64, int64, float64[:], int64, )", cache=True, fastmath=True)
    def sliding_percentile_difference(
        data: np.ndarray,
        upper_pct: int,
        lower_pct: int,
        window_sizes: np.ndarray,
        sample_rate: int,
    ) -> np.ndarray:
        """
        Jitted computes the difference between the upper and lower percentiles within a sliding window for each position
        in the time series using various window sizes. It returns a 2D array where each row corresponds to a position in the time series,
        and each column corresponds to a different window size. The results are calculated as the absolute difference between
        upper and lower percentiles divided by the median of the window.

        :param np.ndarray data: The input time series data.
        :param int upper_pct: The upper percentile value for the window (e.g., 95 for the 95th percentile).
        :param int lower_pct: The lower percentile value for the window (e.g., 5 for the 5th percentile).
        :param np.ndarray window_sizes: An array of window sizes (in seconds) to use for the sliding calculation.
        :param int sample_rate: The sampling rate (samples per second) of the time series data.
        :return np.ndarray: A 2D array containing the difference between upper and lower percentiles for each window size.
        """
        results = np.full((data.shape[0], window_sizes.shape[0]), -1.0)
        for i in prange(window_sizes.shape[0]):
            window_size = int(window_sizes[i] * sample_rate)
            for l, r in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                sample = data[l:r]
                upper_val, lower_val = np.percentile(sample, upper_pct), np.percentile(
                    sample, lower_pct
                )
                median = np.median(sample)
                if median != 0:
                    results[r - 1, i] = np.abs(upper_val - lower_val) / median
                else:
                    results[r - 1, i] = -1.0

        return results.astype(np.float32)

    @staticmethod
    @njit("(float32[:], float64,)", cache=True, fastmath=True)
    def percent_beyond_n_std(data: np.ndarray, n: float) -> float:
        """
        Jitted compute of the ratio of values in time-series more than N standard deviations from the mean of the time-series.

        :parameter np.ndarray data: 1D array representing time-series.
        :parameter float n: Standard deviation cut-off.
        :returns float: Ratio of values in ``data`` that fall more than ``n`` standard deviations from mean of ``data``.

        .. note::
           Adapted from `cesium <https://github.com/cesium-ml/cesium>`_.

        .. image:: _static/img/percent_beyond_n_std.png
           :width: 600
           :align: center

        :examples:
        >>> data = np.array([3.9, 7.5,  4.2, 6.2, 7.5, 3.9, 6.2, 6.5, 7.2, 9.5]).astype(np.float32)
        >>> TimeseriesFeatureMixin().percent_beyond_n_std(data=data, n=1)
        >>> 0.1

        """

        target = (np.std(data) * n) + np.mean(data)
        return np.argwhere(np.abs(data) > target).shape[0] / data.shape[0]

    @staticmethod
    @njit("(float32[:], float64, float64[:], int64,)", cache=True, fastmath=True)
    def sliding_percent_beyond_n_std(
        data: np.ndarray, n: float, window_sizes: np.ndarray, sample_rate: int
    ) -> np.ndarray:
        """
        Computed the percentage of data points that exceed 'n' standard deviations from the mean for each position in
        the time series using various window sizes. It returns a 2D array where each row corresponds to a position in the time series,
        and each column corresponds to a different window size. The results are given as a percentage of data points beyond the threshold.

        :param np.ndarray data: The input time series data.
        :param float n: The number of standard deviations to determine the threshold.
        :param np.ndarray window_sizes: An array of window sizes (in seconds) to use for the sliding calculation.
        :param int sample_rate: The sampling rate (samples per second) of the time series data.
        :return np.ndarray: A 2D array containing the percentage of data points beyond the specified 'n' standard deviations for each window size.
        """

        results = np.full((data.shape[0], window_sizes.shape[0]), -1.0)
        target = (np.std(data) * n) + np.mean(data)
        for i in prange(window_sizes.shape[0]):
            window_size = int(window_sizes[i] * sample_rate)
            for l, r in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                sample = data[l:r]
                results[r - 1, i] = (
                    np.argwhere(np.abs(sample) > target).shape[0] / sample.shape[0]
                )

        return results.astype(np.float32)

    @staticmethod
    @njit("(float32[:], int64, int64, )", cache=True, fastmath=True)
    def percent_in_percentile_window(data: np.ndarray, upper_pct: int, lower_pct: int):
        """
        Jitted compute of the ratio of values in time-series that fall between the ``upper`` and ``lower`` percentile.

        :parameter np.ndarray data: 1D array of representing time-series.
        :parameter int upper_pct: Upper-boundary percentile.
        :parameter int lower_pct: Lower-boundary percentile.
        :returns float: Ratio of values in ``data`` that fall within ``upper_pct`` and ``lower_pct`` percentiles.

        .. note::
           Adapted from `cesium <https://github.com/cesium-ml/cesium>`_.

        .. image:: _static/img/percent_in_percentile_window.png
           :width: 600
           :align: center

        :example:
        >>> data = np.array([3.9, 7.5,  4.2, 6.2, 7.5, 3.9, 6.2, 6.5, 7.2, 9.5]).astype(np.float32)
        >>> TimeseriesFeatureMixin().percent_in_percentile_window(data, upper_pct=70, lower_pct=30)
        >>> 0.4
        """

        upper_val, lower_val = np.percentile(data, upper_pct), np.percentile(
            data, lower_pct
        )
        return (
            np.argwhere((data <= upper_val) & (data >= lower_val)).flatten().shape[0]
            / data.shape[0]
        )

    @staticmethod
    @njit("(float32[:], int64, int64, float64[:], int64)", cache=True, fastmath=True)
    def sliding_percent_in_percentile_window(
        data: np.ndarray,
        upper_pct: int,
        lower_pct: int,
        window_sizes: np.ndarray,
        sample_rate: int,
    ):
        """
        Jitted compute of the percentage of data points falling within a percentile window in a sliding manner.

        The function computes the percentage of data points within the specified percentile window for each position in the time series
        using various window sizes. It returns a 2D array where each row corresponds to a position in the time series, and each column
        corresponds to a different window size. The results are given as a percentage of data points within the percentile window.

        :param np.ndarray data: The input time series data.
        :param int upper_pct: The upper percentile value for the window (e.g., 95 for the 95th percentile).
        :param int lower_pct: The lower percentile value for the window (e.g., 5 for the 5th percentile).
        :param np.ndarray window_sizes: An array of window sizes (in seconds) to use for the sliding calculation.
        :param int sample_rate: The sampling rate (samples per second) of the time series data.
        :return np.ndarray: A 2D array containing the percentage of data points within the percentile window for each window size.

        """
        results = np.full((data.shape[0], window_sizes.shape[0]), -1.0)
        upper_val, lower_val = np.percentile(data, upper_pct), np.percentile(
            data, lower_pct
        )
        for i in prange(window_sizes.shape[0]):
            window_size = int(window_sizes[i] * sample_rate)
            for l, r in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                sample = data[l:r]
                results[r - 1, i] = (
                    np.argwhere((sample <= upper_val) & (sample >= lower_val))
                    .flatten()
                    .shape[0]
                    / sample.shape[0]
                )

        return results.astype(np.float32)

    @staticmethod
    @njit("(float32[:],)", fastmath=True, cache=True)
    def petrosian_fractal_dimension(data: np.ndarray) -> float:
        """
        Calculate the Petrosian Fractal Dimension (PFD) of a given time series data. The PFD is a measure of the
        irregularity or self-similarity of a time series. Larger values indicate higher complexity. Lower values indicate lower complexity.

        .. note::
           The PFD is computed based on the number of sign changes in the first derivative of the time series. If the input data is empty or no sign changes are found, the PFD is returned as -1.0. Adapted from `eeglib <https://github.com/Xiul109/eeglib/>`_.

        .. math::
           PFD = \\frac{\\log_{10}(N)}{\\log_{10}(N) + \\log_{10}\\left(\\frac{N}{N + 0.4 \\cdot zC}\\right)}

        :parameter np.ndarray data: A 1-dimensional numpy array containing the time series data.
        :returns float: The Petrosian Fractal Dimension of the input time series.

        :examples:
        >>> t = np.linspace(0, 50, int(44100 * 2.0), endpoint=False)
        >>> sine_wave = 1.0 * np.sin(2 * np.pi * 1.0 * t).astype(np.float32)
        >>> TimeseriesFeatureMixin().petrosian_fractal_dimension(data=sine_wave)
        >>> 1.0000398187022719
        >>> np.random.shuffle(sine_wave)
        >>> TimeseriesFeatureMixin().petrosian_fractal_dimension(data=sine_wave)
        >>> 1.0211625348743218
        """

        data = (data - np.min(data)) / (np.max(data) - np.min(data))
        derivative = data[1:] - data[:-1]
        if derivative.shape[0] == 0:
            return -1.0
        zC, last_val = 0, -1
        if derivative[0] > 0.0:
            last_val = 1
        for i in prange(1, derivative.shape[0]):
            current_val = -1
            if derivative[i] > 0.0:
                current_val = 1
            if last_val != current_val:
                zC += 1
            last_val = current_val
        if zC == 0:
            return -1.0

        return np.log10(data.shape[0]) / (
            np.log10(data.shape[0])
            + np.log10(data.shape[0] / (data.shape[0] + 0.4 * zC))
        )

    @staticmethod
    @njit("(float32[:], float64[:], int64)", fastmath=True, cache=True)
    def sliding_petrosian_fractal_dimension(
        data: np.ndarray, window_sizes: np.ndarray, sample_rate: int
    ) -> np.ndarray:
        """
        Jitted compute of Petrosian Fractal Dimension over sliding windows in a data array.

        This method computes the Petrosian Fractal Dimension for sliding windows of varying sizes applied
        to the input data array. The Petrosian Fractal Dimension is a measure of signal complexity.

        :param np.ndarray data: Input data array.
        :param np.ndarray window_sizes: Array of window sizes (in seconds).
        :param int sample_rate: Sampling rate of the data in samples per second.
        :return np.ndarray: An array containing Petrosian Fractal Dimension values for each window size and data
                            point. The shape of the result array is (data.shape[0], window_sizes.shape[0]).

        """

        results = np.full((data.shape[0], window_sizes.shape[0]), -1.0)
        for i in prange(window_sizes.shape[0]):
            window_size = int(window_sizes[i] * sample_rate)
            for l, r in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                sample = (data[l:r] - np.min(data[l:r])) / (
                    np.max(data[l:r]) - np.min(data[l:r])
                )
                derivative = sample[1:] - sample[:-1]
                if derivative.shape[0] == 0:
                    results[r - 1, i] = -1.0
                else:
                    zC, last_val = 0, -1
                    if derivative[0] > 0.0:
                        last_val = 1
                    for j in prange(1, derivative.shape[0]):
                        current_val = -1
                        if derivative[j] > 0.0:
                            current_val = 1
                        if last_val != current_val:
                            zC += 1
                        last_val = current_val
                    if zC == 0:
                        results[r - 1, i] = -1.0
                    else:
                        results[r - 1, i] = np.log10(sample.shape[0]) / (
                            np.log10(sample.shape[0])
                            + np.log10(sample.shape[0] / (sample.shape[0] + 0.4 * zC))
                        )

        return results.astype(np.float32)

    @staticmethod
    @njit("(float32[:], int64)")
    def higuchi_fractal_dimension(data: np.ndarray, kmax: int = 10):
        """
        Jitted compute of the Higuchi Fractal Dimension of a given time series data. The Higuchi Fractal Dimension provides a measure of the fractal
        complexity of a time series.

        The maximum value of k used in the calculation. Increasing kmax considers longer sequences
            of data, providing a more detailed analysis of fractal complexity. Default is 10.

        :parameter np.ndarray data: A 1-dimensional numpy array containing the time series data.
        :parameter int kmax: The maximum value of k used in the calculation. Increasing kmax considers longer sequences of data, providing a more detailed analysis of fractal complexity. Default is 10.
        :returns float: The Higuchi Fractal Dimension of the input time series.

        .. note::
           - Adapted from `eeglib <https://github.com/Xiul109/eeglib/>`_.

        .. math::
           HFD = \\frac{\\log(N)}{\\log(N) + \\log\\left(\\frac{N}{N + 0.4 \\cdot zC}\\right)}

        :example:
        >>> t = np.linspace(0, 50, int(44100 * 2.0), endpoint=False)
        >>> sine_wave = 1.0 * np.sin(2 * np.pi * 1.0 * t).astype(np.float32)
        >>> sine_wave = (sine_wave - np.min(sine_wave)) / (np.max(sine_wave) - np.min(sine_wave))
        >>> TimeseriesFeatureMixin().higuchi_fractal_dimension(data=data, kmax=10)
        >>> 1.0001506805419922
        >>> np.random.shuffle(sine_wave)
        >>> TimeseriesFeatureMixin().higuchi_fractal_dimension(data=data, kmax=10)
        >>> 1.9996402263641357
        """

        L, N = np.zeros(kmax - 1), len(data)
        x = np.hstack(
            (
                -np.log(np.arange(2, kmax + 1)).reshape(-1, 1).astype(np.float32),
                np.ones(kmax - 1).reshape(-1, 1).astype(np.float32),
            )
        )
        for k in prange(2, kmax + 1):
            Lk = np.zeros(k)
            for m in range(0, k):
                Lmk = 0
                for i in range(1, (N - m) // k):
                    Lmk += abs(data[m + i * k] - data[m + i * k - k])
                denominator = ((N - m) // k) * k * k
                if denominator == 0:
                    return -1
                Lk[m] = Lmk * (N - 1) / (((N - m) // k) * k * k)
            Laux = np.mean(Lk)
            Laux = 0.01 / k if Laux == 0 else Laux
            L[k - 2] = np.log(Laux)

        return np.linalg.lstsq(x, L.astype(np.float32))[0][0]

    @staticmethod
    @njit("(float32[:], int64, int64,)", fastmath=True)
    def permutation_entropy(data: np.ndarray, dimension: int, delay: int) -> float:
        """
        Calculate the permutation entropy of a time series.

        Permutation entropy is a measure of the complexity of a time series data by quantifying
        the irregularity and unpredictability of its order patterns. It is computed based on the
        frequency of unique order patterns of a given dimension in the time series data.

        The permutation entropy (PE) is calculated using the following formula:

        .. math::
            PE = - \\sum(p_i \\log(p_i))

        where:
        - PE is the permutation entropy.
        - p_i is the probability of each unique order pattern.

        :param numpy.ndarray data: The time series data for which permutation entropy is calculated.
        :param int dimension: It specifies the length of the order patterns to be considered.
        :param int delay: Time delay between elements in an order pattern.
        :return float: The permutation entropy of the time series, indicating its complexity and predictability. A higher permutation entropy value indicates higher complexity and unpredictability in the time series.

        :example:
        >>> t = np.linspace(0, 50, int(44100 * 2.0), endpoint=False)
        >>> sine_wave = 1.0 * np.sin(2 * np.pi * 1.0 * t).astype(np.float32)
        >>> TimeseriesFeatureMixin().permutation_entropy(data=sine_wave, dimension=3, delay=1)
        >>> 0.701970058666407
        >>> np.random.shuffle(sine_wave)
        >>> TimeseriesFeatureMixin().permutation_entropy(data=sine_wave, dimension=3, delay=1)
        >>> 1.79172449934604
        """

        n, permutations, counts = len(data), List(), List()
        for i in prange(n - (dimension - 1) * delay):
            indices = np.arange(i, i + dimension * delay, delay)
            permutation = List(np.argsort(data[indices]))
            is_unique = True
            for j in range(len(permutations)):
                p = permutations[j]
                if len(p) == len(permutation):
                    is_equal = True
                    for k in range(len(p)):
                        if p[k] != permutation[k]:
                            is_equal = False
                            break
                    if is_equal:
                        is_unique = False
                        counts[j] += 1
                        break
            if is_unique:
                permutations.append(permutation)
                counts.append(1)

        total_permutations = len(permutations)
        probs = np.empty(total_permutations, dtype=types.float64)
        for i in prange(total_permutations):
            probs[i] = counts[i] / (n - (dimension - 1) * delay)

        return -np.sum(probs * np.log(probs))

    @staticmethod
    @njit("(float32[:],)", fastmath=True)
    def line_length(data: np.ndarray) -> float:
        """
        Calculate the line length of a 1D array.

        Line length is a measure of signal complexity and is computed by summing the absolute
        differences between consecutive elements of the input array. Used in EEG
        analysis and other signal processing applications to quantify variations in the signal.

        :param numpy.ndarray data: The 1D array for which the line length is to be calculated.
        :return float: The line length of the input array, indicating its complexity.

        .. math::

            LL = \sum_{i=1}^{N-1} |x[i] - x[i-1]|

        where:
        - LL is the line length.
        - N is the number of elements in the input data array.
        - x[i] represents the value of the data at index i.


        .. image:: _static/img/line_length.png
           :width: 600
           :align: center

        :example:
        >>> data = np.array([1, 4, 2, 3, 5, 6, 8, 7, 9, 10]).astype(np.float32)
        >>> TimeseriesFeatureMixin().line_length(data=data)
        >>> 12.0
        """

        diff = np.abs(np.diff(data.astype(np.float64)))
        return np.sum(diff[1:])

    @staticmethod
    @njit("(float32[:], float64[:], int64)", fastmath=True)
    def sliding_line_length(
        data: np.ndarray, window_sizes: np.ndarray, sample_rate: int
    ) -> np.ndarray:
        """
        Jitted compute of  sliding line length for a given time series using different window sizes.

        The function computes line length for the input data using various window sizes. It returns a 2D array where each row
        corresponds to a position in the time series, and each column corresponds to a different window size. The line length
        is calculated for each window, and the results are returned as a 2D array of float32 values.

        :param np.ndarray data: 1D array input data.
        :param window_sizes: An array of window sizes (in seconds) to use for line length calculation.
        :param sample_rate: The sampling rate (samples per second) of the time series data.
        :return np.ndarray: A 2D array containing line length values for each window size at each position in the time series.

        .. image:: _static/img/sliding_line_length.png
           :width: 600
           :align: center

        :examples:
        >>> data = np.array([1, 4, 2, 3, 5, 6, 8, 7, 9, 10]).astype(np.float32)
        >>> TimeseriesFeatureMixin().sliding_line_length(data=data, window_sizes=np.array([1.0]), sample_rate=2)
        """

        results = np.full((data.shape[0], window_sizes.shape[0]), -1.0)
        for i in prange(window_sizes.shape[0]):
            window_size = int(window_sizes[i] * sample_rate)
            for l, r in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                sample = data[l:r]
                results[r - 1, i] = np.sum(np.abs(np.diff(sample.astype(np.float64))))
        return results.astype(np.float32)

    @staticmethod
    @njit("(float32[:], float64[:], int64)", fastmath=True, cache=True)
    def sliding_variance(
        data: np.ndarray, window_sizes: np.ndarray, sample_rate: int
    ) -> np.ndarray:
        """
        Jitted compute of the variance of data within sliding windows of varying sizes applied to
        the input data array. Variance is a measure of data dispersion or spread.

        .. image:: _static/img/sliding_variance.png
           :width: 600
           :align: center

        :param data: 1d input data array.
        :param window_sizes: Array of window sizes (in seconds).
        :param sample_rate: Sampling rate of the data in samples per second.
        :return: Variance values for each window size and data point. The shape of the result array is (data.shape[0], window_sizes.shape[0]).

        :example:
        >>> data = np.array([1, 2, 3, 1, 2, 9, 17, 2, 10, 4]).astype(np.float32)
        >>> TimeseriesFeatureMixin().sliding_variance(data=data, window_sizes=np.array([0.5]), sample_rate=10)
        >>> [[-1.],[-1.],[-1.],[-1.],[ 0.56],[ 8.23],[35.84],[39.20],[34.15],[30.15]])
        """

        results = np.full((data.shape[0], window_sizes.shape[0]), -1.0)
        for i in prange(window_sizes.shape[0]):
            window_size = int(window_sizes[i] * sample_rate)
            for l, r in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                sample = data[l:r]
                results[r - 1, i] = np.var(sample)

        return results

    @staticmethod
    @njit(
        "(float32[:], float64[:], int64, types.ListType(types.unicode_type))",
        fastmath=True,
        cache=True,
    )
    def sliding_descriptive_statistics(
        data: np.ndarray,
        window_sizes: np.ndarray,
        sample_rate: int,
        statistics: Literal[
            "var",
            "max",
            "min",
            "std",
            "median",
            "mean",
            "mad",
            "sum",
            "mac",
            "rms",
            "absenergy",
        ],
    ) -> np.ndarray:
        """
        Jitted compute of descriptive statistics over sliding windows in 1D data array.

        Computes various descriptive statistics (e.g., variance, maximum, minimum, standard deviation,
        median, mean, median absolute deviation) for sliding windows of varying sizes applied to the input data array.

        :param np.ndarray data: 1D input data array.
        :param np.ndarray window_sizes: Array of window sizes (in seconds).
        :param int sample_rate: Sampling rate of the data in samples per second.
        :param types.ListType(types.unicode_type) statistics: List of statistics to compute. Options: 'var', 'max', 'min', 'std', 'median', 'mean', 'mad', 'sum', 'mac', 'rms', 'abs_energy'.
        :return np.ndarray: Array containing the selected descriptive statistics for each window size, data point, and statistic type. The shape of the result array is (len(statistics), data.shape[0], window_sizes.shape[0).

        .. note::
           - The `statistics` parameter should be a list containing one or more of the following statistics:
           'var' (variance), 'max' (maximum), 'min' (minimum), 'std' (standard deviation), 'median' (median),
           'mean' (mean), 'mad' (median absolute deviation), 'sum' (sum), 'mac' (mean absolute change),
           'rms' (root mean square), 'absenergy' (absolute energy).
           - If the statistics list is ['var', 'max', 'mean'], the
           3rd dimension order in the result array will be: [variance, maximum, mean]

        :example:
        >>> data = np.array([1, 4, 2, 3, 5, 6, 8, 7, 9, 10]).astype(np.float32)
        >>> results = TimeseriesFeatureMixin().sliding_descriptive_statistics(data=data, window_sizes=np.array([1.0, 5.0]), sample_rate=2, statistics=typed.List(['var', 'max']))
        """

        results = np.full((len(statistics), data.shape[0], window_sizes.shape[0]), -1.0)
        for j in prange(len(statistics)):
            for i in prange(window_sizes.shape[0]):
                window_size = int(window_sizes[i] * sample_rate)
                for l, r in zip(
                    prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
                ):
                    sample = data[l:r]
                    if statistics[j] == "var":
                        results[j, r - 1, i] = np.var(sample)
                    elif statistics[j] == "max":
                        results[j, r - 1, i] = np.max(sample)
                    elif statistics[j] == "min":
                        results[j, r - 1, i] = np.min(sample)
                    elif statistics[j] == "std":
                        results[j, r - 1, i] = np.std(sample)
                    elif statistics[j] == "median":
                        results[j, r - 1, i] = np.median(sample)
                    elif statistics[j] == "mean":
                        results[j, r - 1, i] = np.mean(sample)
                    elif statistics[j] == "mad":
                        results[j, r - 1, i] = np.median(
                            np.abs(sample - np.median(sample))
                        )
                    elif statistics[j] == "sum":
                        results[j, r - 1, i] = np.sum(sample)
                    elif statistics[j] == "mac":
                        results[j, r - 1, i] = np.mean(np.abs(sample[1:] - sample[:-1]))
                    elif statistics[j] == "rms":
                        results[j, r - 1, i] = np.sqrt(np.mean(sample**2))
                    elif statistics[j] == "absenergy":
                        results[j, r - 1, i] = np.sqrt(np.sum(sample**2))

        return results.astype(np.float32)

    @staticmethod
    def dominant_frequencies(
        data: np.ndarray,
        fps: float,
        k: int,
        window_function: Literal["Hann", "Hamming", "Blackman"] = None,
    ):
        """Find the K dominant frequencies within a feature vector"""

        if window_function == "Hann":
            data = data * np.hanning(len(data))
        elif window_function == "Hamming":
            data = data * np.hamming(len(data))
        elif window_function == "Blackman":
            data = data * np.blackman(len(data))
        fft_result = np.fft.fft(data)
        frequencies = np.fft.fftfreq(data.shape[0], 1 / fps)
        magnitude = np.abs(fft_result)

        return frequencies[np.argsort(magnitude)[-(k + 1) : -1]]

    @staticmethod
    @njit(
        [
            (float32[:], float64, boolean),
            (float32[:], float64, types.misc.Omitted(True)),
        ]
    )
    def longest_strike(data: np.ndarray, threshold: float, above: bool = True) -> int:
        """
        Jitted compute of the length of the longest consecutive sequence of values in the input data that either exceed
        or fall below a specified threshold.

        .. image:: _static/img/longest_strike.png
           :width: 700
           :align: center

        :param np.ndarray data: The input 1D NumPy array containing the values to be analyzed.
        :param float threshold: The threshold value used for the comparison.
        :param bool above: If True, the function looks for strikes where values are above or equal to the threshold. If False, it looks for strikes where values are below or equal to the threshold.
        :return int: The length of the longest strike that satisfies the condition.

        :example:
        >>> data = np.array([1, 8, 2, 10, 8, 6, 8, 1, 1, 1]).astype(np.float32)
        >>> TimeseriesFeatureMixin().longest_strike(data=data, threshold=7, above=True)
        >>> 2
        >>> TimeseriesFeatureMixin().longest_strike(data=data, threshold=7, above=False)
        >>> 3
        """

        result, l, r, cnt = -np.inf, 0, 0, 0
        while l < data.shape[0]:
            if above:
                if data[l] >= threshold:
                    cnt, r = cnt + 1, r + 1
                    while data[r] >= threshold and r < data.shape[0]:
                        cnt, r = cnt + 1, r + 1
            else:
                if data[l] <= threshold:
                    cnt, r = cnt + 1, r + 1
                    while data[r] <= threshold and r < data.shape[0]:
                        cnt, r = cnt + 1, r + 1

            l += 1
            if cnt > result:
                result = cnt
            if data.shape[0] - l < result:
                break
            r, cnt = l, 0
        return result

    @staticmethod
    @njit(
        [
            (float32[:], float64, float64[:], int64, boolean),
            (float32[:], float64, float64[:], int64, types.misc.Omitted(True)),
        ]
    )
    def sliding_longest_strike(
        data: np.ndarray,
        threshold: float,
        time_windows: np.ndarray,
        sample_rate: int,
        above: bool,
    ) -> np.ndarray:
        """
        Jitted compute of the length of the longest strike of values within sliding time windows that satisfy a given condition.

        Calculates the length of the longest consecutive sequence of values in a 1D NumPy array, where each
        sequence is determined by a sliding time window. The condition is specified by a threshold, and
        you can choose whether to look for values above or below the threshold.

        .. image:: _static/img/sliding_longest_strike.png
           :width: 700
           :align: center

        :param np.ndarray data: The input 1D NumPy array containing the values to be analyzed.
        :param float threshold: The threshold value used for the comparison.
        :param np.ndarray time_windows: An array containing the time window sizes in seconds.
        :param int sample_rate: The sample rate in samples per second.
        :param bool above: If True, the function looks for strikes where values are above or equal to the threshold. If False, it looks for strikes where values are below or equal to the threshold.
        :return np.ndarray: A 2D NumPy array with dimensions (data.shape[0], time_windows.shape[0]). Each element in the array represents the length of the longest strike that satisfies the condition for the
        corresponding time window.

        :example:
        >>> data = np.array([1, 8, 2, 10, 8, 6, 8, 1, 1, 1]).astype(np.float32)
        >>> TimeseriesFeatureMixin().sliding_longest_strike(data=data, threshold=7, above=True, time_windows=np.array([1.0]), sample_rate=2)
        >>> [[-1.][ 1.][ 1.][ 1.][ 2.][ 1.][ 1.][ 1.][ 0.][ 0.]]
        >>> TimeseriesFeatureMixin().sliding_longest_strike(data=data, threshold=7, above=True, time_windows=np.array([1.0]), sample_rate=2)
        >>> [[-1.][ 1.][ 1.][ 1.][ 0.][ 1.][ 1.][ 1.][ 2.][ 2.]]
        """

        results = np.full((data.shape[0], time_windows.shape[0]), -1.0)

        for i in prange(time_windows.shape[0]):
            window_size = int(time_windows[i] * sample_rate)
            for l1, r1 in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                sample = data[l1:r1]
                result, l, r, cnt = -np.inf, 0, 0, 0

                while l < sample.shape[0]:
                    if above:
                        if sample[l] >= threshold:
                            cnt, r = cnt + 1, r + 1
                            while sample[r] >= threshold and r < sample.shape[0]:
                                cnt, r = cnt + 1, r + 1
                    else:
                        if sample[l] <= threshold:
                            cnt, r = cnt + 1, r + 1
                            while sample[r] <= threshold and r < sample.shape[0]:
                                cnt, r = cnt + 1, r + 1

                    l += 1
                    if cnt > result:
                        result = cnt
                    if data.shape[0] - l < result:
                        results[r - 1, i] = result
                        break
                    r, cnt = l, 0

                results[r1 - 1, i] = result

            return results

    @staticmethod
    @njit(
        [
            (float32[:], float64, int64, boolean),
            (float32[:], float64, int64, types.misc.Omitted(True)),
        ]
    )
    def time_since_previous_threshold(
        data: np.ndarray, threshold: float, sample_rate: int, above: bool
    ) -> np.ndarray:
        """
        Jitted compute of the time (in seconds) that has elapsed since the last occurrence of a value above (or below)
        a specified threshold in a time series. The time series is assumed to have a constant sample rate.

        .. image:: _static/img/time_since_previous_threshold.png
           :width: 600
           :align: center

        :param np.ndarray data: The input 1D array containing the time series data.
        :param int threshold: The threshold value used for the comparison.
        :param int sample_rate: The sample rate of the time series in samples per second.
        :param bool above: If True, the function looks for values above or equal to the threshold. If False, it looks for values below or equal to the threshold.
        :return np.ndarray: A 1D array of the same length as the input data. Each element represents the time elapsed (in seconds) since the last occurrence of the threshold value. If no threshold value is found before the current data point, the corresponding result is set to -1.0.

        :examples:
        >>> data = np.array([1, 8, 2, 10, 8, 6, 8, 1, 1, 1]).astype(np.float32)
        >>> TimeseriesFeatureMixin().time_since_previous_threshold(data=data, threshold=7.0, above=True, sample_rate=2.0)
        >>> [-1. ,  0. ,  0.5,  0. ,  0. ,  0.5,  0. ,  0.5,  1. ,  1.5]
        >>> TimeseriesFeatureMixin().time_since_previous_threshold(data=data, threshold=7.0, above=False, sample_rate=2.0)
        >>> [0. , 0.5, 0. , 0.5, 1. , 0. , 0.5, 0. , 0. , 0. ]
        """

        results = np.full((data.shape[0]), -1.0)
        if above:
            criterion_idx = np.argwhere(data > threshold).flatten()
        else:
            criterion_idx = np.argwhere(data < threshold).flatten()

        for i in prange(data.shape[0]):
            if above and (data[i] > threshold):
                results[i] = 0.0
            elif not above and (data[i] < threshold):
                results[i] = 0.0
            else:
                x = criterion_idx[np.argwhere(criterion_idx < i).flatten()]
                if len(x) > 0:
                    results[i] = (i - x[-1]) / sample_rate

        return results

    @staticmethod
    @njit(
        [
            (float32[:], float64, int64, boolean),
            (float32[:], float64, int64, types.misc.Omitted(True)),
        ]
    )
    def time_since_previous_target_value(
        data: np.ndarray, value: float, sample_rate: int, inverse: bool = False
    ) -> np.ndarray:
        """
        Calculate the time duration (in seconds) since the previous occurrence of a specific value in a data array.

        Calculates the time duration, in seconds, between each data point and the previous occurrence
        of a specific value within the data array.

        .. image:: _static/img/time_since_previous_target_value.png
           :width: 700
           :align: center

        :param np.ndarray data: The input 1D array containing the time series data.
        :param float value: The specific value to search for in the data array.
        :param int sample_rate: The sampling rate which data points were collected. It is used to calculate the time duration in seconds.
        :param bool inverse: If True, the function calculates the time since the previous value that is NOT equal to the specified 'value'. If False, it calculates the time since the previous occurrence of the specified 'value'.
        :returns np.ndarray: A 1D NumPy array containing the time duration (in seconds) since the previous occurrence of the specified 'value' for each data point.

        :example:
        >>> data = np.array([8, 8, 2, 10, 8, 6, 8, 1, 1, 1]).astype(np.float32)
        >>> TimeseriesFeatureMixin().time_since_previous_target_value(data=data, value=8.0, inverse=False, sample_rate=2.0)
        >>> [0. , 0. , 0.5, 1. , 0. , 0.5, 0. , 0.5, 1. , 1.5])
        >>> TimeseriesFeatureMixin().time_since_previous_target_value(data=data, value=8.0, inverse=True, sample_rate=2.0)
        >>> [-1. , -1. ,  0. ,  0. ,  0.5,  0. ,  0.5,  0. ,  0. ,  0. ]
        """

        results = np.full((data.shape[0]), -1.0)
        if not inverse:
            criterion_idx = np.argwhere(data == value).flatten()
        else:
            criterion_idx = np.argwhere(data != value).flatten()
        if criterion_idx.shape[0] == 0:
            return np.full((data.shape[0]), -1.0)
        for i in prange(data.shape[0]):
            if not inverse and (data[i] == value):
                results[i] = 0
            elif inverse and (data[i] != value):
                results[i] = 0
            else:
                x = criterion_idx[np.argwhere(criterion_idx < i).flatten()]
                if len(x) > 0:
                    results[i] = (i - x[-1]) / sample_rate
        return results

    @staticmethod
    @njit("(float32[:],)")
    def benford_correlation(data: np.ndarray) -> float:
        """
        Jitted compute of the correlation between the Benford's Law distribution and the first-digit distribution of given data.

        Benford's Law describes the expected distribution of leading (first) digits in many real-life datasets. This function
        calculates the correlation between the expected Benford's Law distribution and the actual distribution of the
        first digits in the provided data.

        .. image:: _static/img/benford_correlation.png
           :width: 600
           :align: center

        .. note::
           Adapted from `tsfresh <https://tsfresh.readthedocs.io/en/latest/_modules/tsfresh/feature_extraction/feature_calculators.html#benford_correlation>`_.

           The returned correlation values are calculated using Pearson's correlation coefficient.

        :param np.ndarray data: The input 1D array containing the time series data.
        :return float: The correlation coefficient between the Benford's Law distribution and the first-digit distribution in the input data. A higher correlation value suggests that the data follows the expected distribution more closely.

        :examples:
        >>> data = np.array([1, 8, 2, 10, 8, 6, 8, 1, 1, 1]).astype(np.float32)
        >>> TimeseriesFeatureMixin().benford_correlation(data=data)
        >>> 0.6797500374831786
        """

        data = np.abs(data)
        benford_distribution = np.array([np.log10(1 + 1 / n) for n in range(1, 10)])
        first_vals, digit_ratio = np.full((data.shape[0]), np.nan), np.full(9, np.nan)
        for i in prange(data.shape[0]):
            first_vals[i] = data[i] // 10 ** (int(np.log10(data[i])) - 1 + 1)

        for i in range(1, 10):
            digit_ratio[i - 1] = np.argwhere(first_vals == i).shape[0] / data.shape[0]

        return np.corrcoef(benford_distribution, digit_ratio)[0, 1]

    @staticmethod
    @njit("(float32[:], float64[:], int64)")
    def sliding_benford_correlation(
        data: np.ndarray, time_windows: np.ndarray, sample_rate: int
    ) -> np.ndarray:
        """
        Calculate the sliding Benford's Law correlation coefficient for a given dataset within
        specified time windows.

        Benford's Law is a statistical phenomenon where the leading digits of many datasets follow a
        specific distribution pattern. This function calculates the correlation between the observed
        distribution of leading digits in a dataset and the ideal Benford's Law distribution.

        .. note::
           Adapted from `tsfresh <https://tsfresh.readthedocs.io/en/latest/_modules/tsfresh/feature_extraction/feature_calculators.html#benford_correlation>`_.

           The returned correlation values are calculated using Pearson's correlation coefficient.

        The correlation coefficient is calculated between the observed leading digit distribution and
        the ideal Benford's Law distribution.

        .. math::

            P(d) = \\log_{10}\\left(1 + \\frac{1}{d}\\right) \\quad \\text{for } d \\in \{1, 2, \\ldots, 9\}

        :param np.ndarray data: The input 1D array containing the time series data.
        :param np.ndarray time_windows: A 1D array containing the time windows (in seconds) for which the correlation will be calculated at different points in the dataset.
        :param int sample_rate: The sample rate, indicating how many data points are collected per second.
        :return np.ndarray: 2D array containing the correlation coefficient values for each time window. With time window lenths represented by different columns.

        :examples:
        >>> data = np.array([1, 8, 2, 10, 8, 6, 8, 1, 1, 1]).astype(np.float32)
        >>> sliding_benford_correlation(data=data, time_windows=np.array([1.0]), sample_rate=2)
        >>> [[ 0.][0.447][0.017][0.877][0.447][0.358][0.358][0.447][0.864][0.864]]
        """

        data = np.abs(data)
        benford_distribution = np.array([np.log10(1 + 1 / n) for n in range(1, 10)])
        results = np.full((data.shape[0], time_windows.shape[0]), 0.0)
        for i in prange(time_windows.shape[0]):
            window_size = int(time_windows[i] * sample_rate)
            for l, r in zip(
                prange(0, data.shape[0] + 1), prange(window_size, data.shape[0] + 1)
            ):
                first_vals, digit_ratio = np.full((data.shape[0]), np.nan), np.full(
                    9, np.nan
                )
                sample = data[l:r]
                for k in range(sample.shape[0]):
                    first_vals[k] = sample[k] // 10 ** (
                        int(np.log10(sample[k])) - 1 + 1
                    )
                for k in range(1, 10):
                    digit_ratio[k - 1] = (
                        np.argwhere(first_vals == k).shape[0] / sample.shape[0]
                    )
                results[r - 1, i] = np.corrcoef(benford_distribution, digit_ratio)[0, 1]

        return results.astype(np.float32)

    @staticmethod
    @njit("(float32[:], float64, float64, float64, float64, float64)", fastmath=True)
    def spike_finder(
        data: np.ndarray,
        sample_rate: int,
        baseline: float,
        min_spike_amplitude: float,
        min_fwhm: float = -np.inf,
        min_half_width: float = -np.inf,
    ) -> float:
        """
        Identify and characterize spikes in a given time-series data sequence. This method identifies spikes in the input data based on the specified criteria and characterizes
        each detected spike by computing its amplitude, full-width at half maximum (FWHM), and half-width.

        :param np.ndarray data: A 1D array containing the input data sequence to analyze.
        :param int sample_rate: The sample rate, indicating how many data points are collected per second.
        :param float baseline: The baseline value used to identify spikes. Any data point above (baseline + min_spike_amplitude) is considered part of a spike.
        :param float min_spike_amplitude: The minimum amplitude (above baseline) required for a spike to be considered.
        :param Optional[float] min_fwhm: The minimum full-width at half maximum (FWHM) for a spike to be included. If not specified, it defaults to negative infinity, meaning it is not considered for filtering.
        :param Optional[float] min_half_width: The minimum half-width required for a spike to be included. If not specified, it defaults to negative infinity, meaning it is not considered for filtering.
        :return tuple: A tuple containing three elements:
            - spike_idx (List[np.ndarray]): A list of 1D arrays, each representing the indices of the data points belonging to a detected spike.
            - spike_vals (List[np.ndarray]): A list of 1D arrays, each containing the values of the data points within a detected spike.
            - spike_dict (Dict[int, Dict[str, float]]): A dictionary where the keys are spike indices, and the values are dictionaries containing spike characteristics including 'amplitude' (spike amplitude), 'fwhm' (FWHM), and 'half_width' (half-width).

        .. notes::
           - The function uses the Numba JIT (Just-In-Time) compilation for optimized performance. Without fastmath=True there is no runtime improvement over standard numpy.

        :example:
        >>> data = np.array([0.1, 0.1, 0.3, 0.1, 10, 10, 8, 0.1, 0.1, 0.1, 10, 10, 8, 99, 0.1, 99, 99, 0.1]).astype(np.float32)
        >>> spike_idx, spike_vals, spike_stats = spike_finder(data=data, baseline=1, min_spike_amplitude=5, sample_rate=2, min_fwhm=-np.inf, min_half_width=0.0002)
        """

        spike_idxs = np.argwhere(data >= baseline + min_spike_amplitude).flatten()
        spike_idx = np.split(
            spike_idxs, np.argwhere(spike_idxs[1:] - spike_idxs[:-1] != 1).flatten() + 1
        )
        spike_dict = Dict.empty(
            key_type=types.int64,
            value_type=Dict.empty(
                key_type=types.unicode_type, value_type=types.float64
            ),
        )
        spike_vals = []
        for i in prange(len(spike_idx)):
            spike_data = data[spike_idx[i]]
            spike_amplitude = np.max(spike_data) - baseline
            half_width_idx = np.argwhere(spike_data > spike_amplitude / 2).flatten()
            spike_dict[i] = {
                "amplitude": np.max(spike_data) - baseline,
                "fwhm": (half_width_idx[-1] - half_width_idx[0]) / sample_rate,
                "half_width": half_width_idx.shape[0] / sample_rate,
            }
            spike_vals.append(spike_data)

        remove_idx = []
        for k, v in spike_dict.items():
            if (v["fwhm"] < min_fwhm) or (v["half_width"] < min_half_width):
                remove_idx.append(k)
        for idx in remove_idx:
            spike_dict.pop(idx)
        spike_idx = [i for j, i in enumerate(spike_idx) if j not in remove_idx]
        spike_vals = [i for j, i in enumerate(spike_vals) if j not in remove_idx]

        return spike_idx, spike_vals, spike_dict

    # @njit("(float32[:], types.List(types.Array(types.int64, 1, 'C')), int64, float64, float64)", fastmath=True)
    @staticmethod
    @jit(nopython=True, fastmath=True)
    def spike_train_finder(
        data: np.ndarray,
        spike_idx: list,
        sample_rate: float,
        min_spike_train_length: float = np.inf,
        max_spike_train_separation: float = np.inf,
    ):
        """
        Identify and analyze spike trains from a list of spike indices.

        This function takes spike indices and additional information, such as the data, sample rate,
        minimum spike train length, and maximum spike train separation, to identify and analyze
        spike trains in the data.

        :param np.ndarray data: The data from which spike trains are extracted.
        :param types.List(types.Array(types.int64, 1, 'C')) data: A list of spike indices, typically as integer timestamps.
        :param float sample_rate: The sample rate of the data.
        :param Optional[float] min_spike_train_length: The minimum length a spike train must have to be considered. Default is set to positive infinity, meaning no minimum length is enforced.
        :param Optional[float] max_spike_train_separation: The maximum allowable separation between spikes in the same train. Default is set to positive infinity, meaning no maximum separation is enforced.
        :return DictType[int64,DictType[unicode_type,float64]]: A dictionary containing information about identified spike trains.

        Each entry in the returned dictionary is indexed by an integer, and contains the following information:
            - 'train_start_time': Start time of the spike train in seconds.
            - 'train_end_time': End time of the spike train in seconds.
            - 'train_start_obs': Start time index in observations.
            - 'train_end_obs': End time index in observations.
            - 'spike_cnt': Number of spikes in the spike train.
            - 'train_length_obs_cnt': Length of the spike train in observations.
            - 'train_length_obs_s': Length of the spike train in seconds.
            - 'train_spike_mean_lengths_s': Mean length of individual spikes in seconds.
            - 'train_spike_std_length_obs': Standard deviation of spike lengths in observations.
            - 'train_spike_std_length_s': Standard deviation of spike lengths in seconds.
            - 'train_spike_max_length_obs': Maximum spike length in observations.
            - 'train_spike_max_length_s': Maximum spike length in seconds.
            - 'train_spike_min_length_obs': Minimum spike length in observations.
            - 'train_spike_min_length_s': Minimum spike length in seconds.
            - 'train_mean_amplitude': Mean amplitude of the spike train.
            - 'train_std_amplitude': Standard deviation of spike amplitudes.
            - 'train_min_amplitude': Minimum spike amplitude.
            - 'train_max_amplitude': Maximum spike amplitude.


        .. note::
           - The function may return an empty dictionary if no spike trains meet the criteria.
           - A required input is ``spike_idx``, which is returned by :func:`~timeseries_features_mixin.TimeseriesFeatureMixin.spike_finder`.

        :example:
        >>> data = np.array([0.1, 0.1, 0.3, 0.1, 10, 10, 8, 0.1, 0.1, 0.1, 10, 10, 8, 99, 0.1, 99, 99, 0.1]).astype(np.float32)
        >>> spike_idx, _, _ = TimeseriesFeatureMixin().spike_finder(data=data, baseline=0.3, min_spike_amplitude=0.2, sample_rate=2, min_fwhm=-np.inf, min_half_width=-np.inf)
        >>> results = TimeseriesFeatureMixin().spike_train_finder(data=data, spike_idx=typed.List(spike_idx), sample_rate=2.0, min_spike_train_length=2.0, max_spike_train_separation=2.0)
        """

        (
            l,
            r,
        ) = (
            0,
            1,
        )
        train_data_idx = []
        train_spikes_idx = []
        max_spike_train_separation = int(max_spike_train_separation * sample_rate)
        min_spike_train_length = int(min_spike_train_length * sample_rate)
        while l < len(spike_idx):
            current_train, current_spike_idx = spike_idx[l], [l]
            while r < len(spike_idx) and (
                (spike_idx[r][0] - current_train[-1]) <= max_spike_train_separation
            ):
                current_train = np.hstack((current_train, spike_idx[r]))
                current_spike_idx.append(r)
                r += 1
            l, r = r, r + 1
            train_data_idx.append(current_train)
            train_spikes_idx.append(current_spike_idx)

        spike_dict = Dict.empty(
            key_type=types.int64,
            value_type=Dict.empty(
                key_type=types.unicode_type, value_type=types.float64
            ),
        )
        for i in prange(len(train_data_idx)):
            if train_data_idx[i].shape[0] >= min_spike_train_length:
                spike_train_amps = data[train_data_idx[i]]
                spike_train_idx = [
                    k for j, k in enumerate(spike_idx) if j in train_spikes_idx[i]
                ]
                train_spike_lengths = np.array(([len(j) for j in spike_train_idx]))
                spike_dict[int(i)] = {
                    "train_start_time": float(train_data_idx[i][0] * sample_rate),
                    "train_end_time": train_data_idx[i][-1] * sample_rate,
                    "train_start_obs": train_data_idx[i][0],
                    "train_end_obs": train_data_idx[i][-1],
                    "spike_cnt": len(train_spikes_idx[i]),
                    "train_length_obs_cnt": len(spike_train_amps),
                    "train_length_obs_s": len(spike_train_amps) / sample_rate,
                    "train_spike_mean_lengths_s": np.mean(train_spike_lengths)
                    * sample_rate,
                    "train_spike_std_length_obs": np.mean(train_spike_lengths),
                    "train_spike_std_length_s": np.mean(train_spike_lengths)
                    * sample_rate,
                    "train_spike_max_length_obs": np.max(train_spike_lengths),
                    "train_spike_max_length_s": np.max(train_spike_lengths)
                    * sample_rate,
                    "train_spike_min_length_obs": np.min(train_spike_lengths),
                    "train_spike_min_length_s": np.min(train_spike_lengths)
                    * sample_rate,
                    "train_mean_amplitude": np.mean(spike_train_amps),
                    "train_std_amplitude": np.std(spike_train_amps),
                    "train_min_amplitude": np.min(spike_train_amps),
                    "train_max_amplitude": np.max(spike_train_amps),
                }

        return spike_dict

    @staticmethod
    def _adf_executor(data: np.ndarray) -> tuple:
        """
        Helper function to execute Augmented Dickey-Fuller (ADF) test on a data segment.
        Called by :meth:`timeseries_features_mixin.TimeseriesFeatureMixin.sliding_stationary_test_test`.
        """

        adfuller_results = adfuller(data)
        return adfuller_results[0], adfuller_results[1]

    @staticmethod
    def _kpss_executor(data: np.ndarray) -> tuple:
        """
        Helper function to execute Kwiatkowski–Phillips–Schmidt–Shin (KPSS) test on a data segment.
        Called by :meth:`timeseries_features_mixin.TimeseriesFeatureMixin.sliding_stationary_test_test`.
        """

        kpss_results = kpss(data)
        return kpss_results[0], kpss_results[1]

    @staticmethod
    def _zivotandrews_executor(data: np.ndarray) -> tuple:
        """
        Helper function to execute Zivot-Andrews structural-break unit-root test on a data segment.
        Called by :meth:`timeseries_features_mixin.TimeseriesFeatureMixin.sliding_stationary_test_test`.
        """
        try:
            za_results = zivot_andrews(data)
            return za_results[0], za_results[1]
        except (np.linalg.LinAlgError, ValueError):
            return 0, 0

    @staticmethod
    def sliding_stationary(
        data: np.ndarray,
        time_windows: np.ndarray,
        sample_rate: int,
        test: Literal["ADF", "KPSS", "ZA"] = "adf",
    ) -> (np.ndarray, np.ndarray):
        """
        Perform the Augmented Dickey-Fuller (ADF), Kwiatkowski-Phillips-Schmidt-Shin (KPSS), or Zivot-Andrews test on sliding windows of time series data.
        Parallel processing using all available cores is used to accelerate computation.

        .. note::
           - ADF: A high p-value suggests non-stationarity, while a low p-value indicates stationarity.
           - KPSS: A high p-value suggests stationarity, while a low p-value indicates non-stationarity.
           - ZA: A high p-value suggests non-stationarity, while a low p-value indicates stationarity.

        :param np.ndarray data: 1-D NumPy array containing the time series data to be tested.
        :param np.ndarray time_windows: A 1-D NumPy array containing the time window sizes in seconds.
        :param np.ndarray sample_rate: The sample rate of the time series data (samples per second).
        :param Literal test: Test to perfrom: Options: 'ADF' (Augmented Dickey-Fuller), 'KPSS' (Kwiatkowski-Phillips-Schmidt-Shin), 'ZA' (Zivot-Andrews).
        :return (np.ndarray, np.ndarray): A tuple of two 2-D NumPy arrays containing test statistics and p-values. - The first array (stat) contains the ADF test statistics. - The second array (p_vals) contains the corresponding p-values

        :example:
        >>> data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        >>> TimeseriesFeatureMixin().sliding_stationary(data=data, time_windows=np.array([2.0]), test='KPSS', sample_rate=2)
        """

        stat = np.full((data.shape[0], time_windows.shape[0]), -1.0)
        p_vals = np.full((data.shape[0], time_windows.shape[0]), -1.0)
        if test == "ADF":
            test_func = TimeseriesFeatureMixin()._adf_executor
        elif test == "KPSS":
            test_func = TimeseriesFeatureMixin()._kpss_executor
        elif test == "ZA":
            test_func = TimeseriesFeatureMixin()._zivotandrews_executor
        else:
            type_hints = get_type_hints(TimeseriesFeatureMixin().sliding_stationary)
            raise InvalidInputError(
                msg=f'Test {test} not recognized. Options: {type_hints["test"]}'
            )
        for i in range(time_windows.shape[0]):
            window_size = int(time_windows[i] * sample_rate)
            strided_data = as_strided(
                data,
                shape=(data.shape[0] - window_size + 1, window_size),
                strides=(data.strides[0] * 1, data.strides[0]),
            )
            with multiprocessing.Pool(find_core_cnt()[0], maxtasksperchild=10) as pool:
                for cnt, result in enumerate(
                    pool.imap(test_func, strided_data, chunksize=1)
                ):
                    stat[cnt + window_size - 1, i] = result[0]
                    p_vals[cnt + window_size - 1, i] = result[1]

        return stat, p_vals

    @staticmethod
    @njit(
        [
            "(float32[:], float64, int64, float64, types.unicode_type)",
            '(float32[:], float64, int64, float64, types.misc.Omitted("mm"))',
            "(float32[:], float64, int64, types.misc.Omitted(1), types.unicode_type)",
            '(float32[:], float64, int64, types.misc.Omitted(1), types.misc.Omitted("mm"))',
        ]
    )
    def acceleration(
        data: np.ndarray,
        pixels_per_mm: float,
        fps: int,
        time_window: float = 1,
        unit: Literal["mm", "cm", "dm", "m"] = "mm",
    ) -> np.ndarray:
        """
        Compute acceleration.

        Computes acceleration from a sequence of body-part coordinates over time. It calculates the difference in velocity between consecutive frames and provides an array of accelerations.

        The computation is based on the formula:

        .. math::

           \\text{{Acceleration}}(t) = \\frac{{\\text{{Norm}}(\\text{{Shift}}(\\text{{data}}[t], t, t-1) - \\text{{data}}[t])}}{{\\text{{pixels\\_per\\_mm}}}}

        where :math:`\\text{{Norm}}` calculates the Euclidean norm, :math:`\\text{{Shift}}(\\text{{array}}, t, t-1)` shifts the array by :math:`t-1` frames, and :math:`\\text{{pixels\\_per\\_mm}}` is the conversion factor from pixels to millimeters.

        .. note::
           By default, acceleration is calculated as change in velocity at millimeters/s. To change the denomitator, modify the ``time_window`` argument. To change the nominator, modify the ``unit`` argument (accepted ``mm``, cm``, ``dm``, ``mm``)

        .. image:: _static/img/acceleration.png
           :width: 700
           :align: center

        :param np.ndarray data: 2D array of size len(frames) x 2 with body-part coordinates.
        :param float pixels_per_mm: Pixels per millimeter of the recorded video.
        :param int fps: Frames per second (FPS) of the recorded video.
        :param float time_window: Rolling time window in seconds. Default is 1.0 representing 1 second.
        :param Literal['mm', 'cm', 'dm', 'm'] unit:  If acceleration should be presented as millimeter, centimeters, decimeter, or meter. Default millimeters.
        :return: Array of accelerations corresponding to each frame.

        :example:
        >>> data = np.array([1, 2, 3, 4, 5, 5, 5, 5, 5, 6]).astype(np.float32)
        >>> TimeseriesFeatureMixin().acceleration(data=data, pixels_per_mm=1.0, fps=2, time_window=1.0)
        >>> [ 0.,  0.,  0.,  0., -1., -1.,  0.,  0.,  1.,  1.]
        """

        results, velocity = np.full((data.shape[0]), 0.0), np.full((data.shape[0]), 0.0)
        size, pv = int(time_window * fps), None
        data_split = np.split(data, list(range(size, data.shape[0], size)))
        for i in range(len(data_split)):
            wS = int(size * i)
            wE = int(wS + size)
            v = np.diff(np.ascontiguousarray(data_split[i]))[0] / pixels_per_mm
            if unit == "cm":
                v = v / 10
            elif unit == "dm":
                v = v / 100
            elif unit == "m":
                v = v / 1000
            if i == 0:
                results[wS:wE] = 0
            else:
                results[wS:wE] = v - pv
            pv = v
        return results


# data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
# TimeseriesFeatureMixin().sliding_stationary_test_test(data=data, time_windows=np.array([2.0]), test='BLAH', sample_rate=2)
#


# data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
# TimeseriesFeatureMixin().sliding_adf_test(data=data, time_windows=np.array([2.0]), sample_rate=2)


# data_sizes = np.array([10, 100, 1000, 10000, 100000, 1000000, 10000000])
# data_sizes = np.array([100000000])
# import time
#
# for j in range(5):
#     print('x')
#     for i in data_sizes:
#         data = np.random.randint(1, 1000, (i,)).astype(np.float32)
#         start = time.time()
#         blah, blah1, bls = TimeseriesFeatureMixin().sliding_hjort_parameters(data=data, window_sizes=np.array([2.0, 4.0]), sample_rate=30)
#         print(time.time() - start)
#
# data_sizes = np.array([1, 10, 100, 1000, 10000, 100000])
# # #data_sizes = np.array([1, 10, 100, 1000, 10000])
# # # data_sizes = np.array([100000000])
# import time
#
# for j in range(5):
#     print('x')
#     for i in data_sizes:
#         data = np.random.randint(1, 1000, (i,)).astype(np.float32)
#         start = time.time()
#         TimeseriesFeatureMixin().sliding_benford_correlation(data=data, time_windows=np.array([1.0, 4.0]), sample_rate=30)
#         print(time.time() - start)


# data = np.array([1, 4, 2, 3, 5, 6, 8, 7, 9, 10]).astype(np.float32)
# results = sliding_descriptive_statistics(data=data, window_sizes=np.array([1.0, 5.0]), sample_rate=2, statistics=typed.List(['mac']))


# t = np.linspace(0, 50, int(44100 * 2.0), endpoint=False)
# sine_wave = 1.0 * np.sin(2 * np.pi * 1.0 * t).astype(np.float32)
# TimeseriesFeatureMixin().petrosian_fractal_dimension(data=sine_wave)
# #1.0000398187022719
# np.random.shuffle(sine_wave)
# TimeseriesFeatureMixin().petrosian_fractal_dimension(data=sine_wave)
# #1.0211625348743218



import pandas as pd
import pytest
import json

from simba.mixins.config_reader import ConfigReader
from simba.utils.read_write import get_fn_ext, read_df, read_config_file


@pytest.fixture(params=['tests/data/test_projects/two_c57/project_folder/project_config.ini'])
def project_config_args(request):
    return request

@pytest.fixture(params=['tests/data/test_projects/two_c57/project_folder/project_config.ini',
                        'tests/data/test_projects/mouse_open_field/project_folder/project_config.ini',
                        'tests/data/test_projects/zebrafish/project_folder/project_config.ini'])
def all_project_config_args(request):
    return request


@pytest.mark.parametrize("config_path, expected_clf_names", [['tests/data/test_projects/two_c57/project_folder/project_config.ini', ['Attack', 'Sniffing']],
                                                            ['tests/data/test_projects/mouse_open_field/project_folder/project_config.ini', ['Attack']],
                                                            ['tests/data/test_projects/zebrafish/project_folder/project_config.ini', ['Rheotaxis']]])
def test_config_reader(config_path, expected_clf_names):
    config_reader = ConfigReader(config_path=config_path)
    config_reader.read_roi_data()
    assert config_reader.get_all_clf_names() == expected_clf_names


def test_get_number_of_header_columns_in_df(project_config_args):
    config_reader = ConfigReader(config_path=project_config_args.param)
    results = config_reader.get_number_of_header_columns_in_df(df=pd.DataFrame(data=[[1, 2, 3], [1, 2, 3]]))
    assert results == 0
    results = config_reader.get_number_of_header_columns_in_df(df=pd.DataFrame(data=[['sdfsdf', 'sdf', 'asd'], ['5', 'asd', 'asd'], [11, 99, 109], [122, 43, 2091]]))
    assert results == 2

@pytest.mark.parametrize("config_path, expected_path",      [['tests/data/test_projects/two_c57/project_folder/project_config.ini', 'Together_1'],
                                                            ['tests/data/test_projects/mouse_open_field/project_folder/project_config.ini', 'Video1'],
                                                            ['tests/data/test_projects/zebrafish/project_folder/project_config.ini', '20200730_AB_7dpf_850nm_0003']])
def test_find_video_of_file(config_path, expected_path):
    config_reader = ConfigReader(config_path=config_path)
    video_path = config_reader.find_video_of_file(video_dir=config_reader.video_dir, filename=expected_path)
    _, video_name, _ = get_fn_ext(video_path)
    assert video_name == expected_path

def test_add_missing_ROI_cols(project_config_args):
    config_reader = ConfigReader(config_path=project_config_args.param)
    results = config_reader.add_missing_ROI_cols(shape_df=pd.DataFrame(data=[[1, 2, 3], [1, 2, 3]], columns=['Name', 'Center_X', 'Center_Y']))
    s = set()
    s.update(['Color BGR', 'Thickness', 'Color name'])
    assert len(s - set(results.columns)) == 0


@pytest.mark.parametrize("config_path, body_part_name, expected_animal_name", [['tests/data/test_projects/two_c57/project_folder/project_config.ini', 'Nose_1', 'simon'],
                                                                               ['tests/data/test_projects/mouse_open_field/project_folder/project_config.ini', 'Left_ear', 'Animal_1'],
                                                                               ['tests/data/test_projects/zebrafish/project_folder/project_config.ini', 'Zebrafish_LeftEye', 'Animal_1']])
def test_find_animal_name_from_body_part_name(config_path, body_part_name, expected_animal_name):
    config_reader = ConfigReader(config_path=config_path)
    animal_name = config_reader.find_animal_name_from_body_part_name(bp_name=body_part_name, bp_dict=config_reader.animal_bp_dict)
    assert animal_name == expected_animal_name

@pytest.mark.parametrize("config_path, expected", [['tests/data/test_projects/two_c57/project_folder/project_config.ini',
                                                    'tests/data/test_projects/two_c57/expected_animal_bp_dict/expected_animal_bp_dict.json']])
def test_create_body_part_dictionary(config_path, expected):
    config_reader = ConfigReader(config_path=config_path)
    bp_dict = config_reader.create_body_part_dictionary(multi_animal_status=config_reader.multi_animal_status,
                                                        animal_id_lst=config_reader.multi_animal_id_list,
                                                        animal_cnt=config_reader.animal_cnt,
                                                        x_cols=config_reader.x_cols,
                                                        y_cols=config_reader.y_cols,
                                                        p_cols=config_reader.p_cols,
                                                        colors=config_reader.clr_lst)
    with open(expected, 'r') as fp:
        assert json.load(fp) == bp_dict

def test_get_body_part_names(project_config_args):
    config_reader = ConfigReader(config_path=project_config_args.param)
    config_reader.get_body_part_names()
    assert len(config_reader.x_cols) == 16
    assert len(config_reader.y_cols) == 16
    assert len(config_reader.y_cols) == 16

def test_drop_bp_cords(project_config_args):
    config_reader = ConfigReader(config_path=project_config_args.param)
    df = read_df(config_reader.machine_results_paths[0], file_type='csv')
    out_df = config_reader.drop_bp_cords(df=df)
    assert (len(df.columns) - len(out_df.columns)) == len(config_reader.bp_headers)


@pytest.mark.parametrize("config_path, expected",      [['tests/data/test_projects/two_c57/project_folder/project_config.ini', 48],
                                                            ['tests/data/test_projects/mouse_open_field/project_folder/project_config.ini', 12],
                                                            ['tests/data/test_projects/zebrafish/project_folder/project_config.ini', 21]])
def test_get_bp_headers(config_path, expected):
    config_reader = ConfigReader(config_path=config_path)
    config_reader.get_bp_headers()
    assert len(config_reader.bp_headers) == expected

@pytest.mark.parametrize("config_path, expected",      [['tests/data/test_projects/two_c57/project_folder/project_config.ini', 2],
                                                            ['tests/data/test_projects/mouse_open_field/project_folder/project_config.ini', 1],
                                                            ['tests/data/test_projects/zebrafish/project_folder/project_config.ini', 1]])
def test_read_config_entry(config_path, expected):
    config = read_config_file(config_path=config_path)
    config_reader = ConfigReader(config_path=config_path)
    result = config_reader.read_config_entry(config=config, section='General settings', option='animal_no', data_type='int')
    assert result == expected

def test_read_video_info_csv(all_project_config_args):
    config_reader = ConfigReader(config_path=all_project_config_args.param)
    df = config_reader.read_video_info_csv(file_path=config_reader.video_info_path)
    print(df)
    assert isinstance(df, pd.DataFrame)

@pytest.mark.parametrize("config_path, video_name, expected_pixels, expected_fps",      [['tests/data/test_projects/two_c57/project_folder/project_config.ini', 'Together_1', 4.0, 30],
                                                                                         ['tests/data/test_projects/mouse_open_field/project_folder/project_config.ini', 'Video1', 2.57425, 30],
                                                                                         ['tests/data/test_projects/zebrafish/project_folder/project_config.ini', '20200730_AB_7dpf_850nm_0003', 1.2273399999999999, 30]])
def test_read_video_info(config_path, video_name, expected_pixels, expected_fps):
    config_reader = ConfigReader(config_path=config_path)
    video_settings, px_per_mm, fps = config_reader.read_video_info(video_name=video_name)
    assert len(video_settings) == 1
    assert px_per_mm == expected_pixels
    assert fps == fps





__author__ = "Simon Nilsson"

import os
import random

import cv2
import numpy as np
import pandas as pd

from simba.data_processors.directing_other_animals_calculator import \
    DirectingOtherAnimalsAnalyzer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.data import create_color_palettes
from simba.utils.enums import Formats
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, get_video_meta_data, read_df
from simba.utils.warnings import NoDataFoundWarning


class DirectingOtherAnimalsVisualizer(ConfigReader, PlottingMixin):
    """
    Visualize when animals are directing towards body-parts of other animals.

    .. important::
       Requires the pose-estimation data for the left ear, right ears and nose of individual animals.
       For better runtime, use :meth:`simba.plotting.Directing_animals_visualizer.DirectingOtherAnimalsVisualizerMultiprocess`.

    .. note::
       `Example of expected output <https://www.youtube.com/watch?v=d6pAatreb1E&list=PLi5Vwf0hhy1R6NDQJ3U28MOUJPfl2YWYl&index=22`_.

        `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-data-tables`>_.

        .. image:: _static/img/directing_other_animals.png
           :width: 450
           :align: center

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str data_path: path to data file
    :parameter dict style_attr: Visualisation attributes (colors and sizes etc.)


    Examples
    -----
    >>> style_attr = {'Show_pose': True,
    >>>               'Pose_circle_size': 3,
    >>>               "Direction_color": 'Random',
    >>>               'Direction_thickness': 4, 'Highlight_endpoints': True}
    >>> _ = DirectingOtherAnimalsVisualizer(config_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini', video_name='Testing_Video_3.mp4', style_attr=style_attr).run()
    """

    def __init__(self, config_path: str, data_path: str, style_attr: dict):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)

        self.data_path = data_path
        _, self.video_name, _ = get_fn_ext(self.data_path)
        self.direction_analyzer = DirectingOtherAnimalsAnalyzer(
            config_path=config_path,
            bool_tables=False,
            summary_tables=False,
            aggregate_statistics_tables=False,
        )
        self.direction_analyzer.outlier_corrected_paths = [data_path]
        self.direction_analyzer.run()
        self.direction_analyzer._transpose_results_to_df()
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.style_attr, self.pose_colors = style_attr, []
        self.colors = get_color_dict()
        if self.style_attr["Show_pose"]:
            self.pose_colors = create_color_palettes(
                self.animal_cnt, int(len(self.x_cols) + 1)
            )
        if self.style_attr["Direction_color"] == "Random":
            self.direction_colors = create_color_palettes(1, int(self.animal_cnt**2))
        else:
            self.direction_colors = [self.colors[self.style_attr["Direction_color"]]]
        self.data_dict = self.direction_analyzer.directionality_df_dict
        self.video_path = self.find_video_of_file(
            video_dir=self.video_dir, filename=self.video_name
        )
        if not os.path.exists(self.directing_animals_video_output_path):
            os.makedirs(self.directing_animals_video_output_path)
        print(f"Processing video {self.video_name}...")

    def run(self):
        """
        Method to create directionality videos. Results are stored in
        the `project_folder/frames/output/ROI_directionality_visualize` directory of the SimBA project

        Returns
        ----------
        None
        """

        self.data_df = read_df(self.data_path, file_type=self.file_type)
        self.video_save_path = os.path.join(
            self.directing_animals_video_output_path, self.video_name + ".mp4"
        )
        self.cap = cv2.VideoCapture(self.video_path)
        self.video_meta_data = get_video_meta_data(self.video_path)
        self.video_data = self.data_dict[self.video_name]
        self.writer = cv2.VideoWriter(
            self.video_save_path,
            self.fourcc,
            self.video_meta_data["fps"],
            (self.video_meta_data["width"], self.video_meta_data["height"]),
        )
        if self.video_name in list(self.video_data["Video"]):
            self.__create_video()
        else:
            NoDataFoundWarning(
                msg=f"SimBA skipping video {self.video_name}: No animals are directing each other in the video."
            )
        self.timer.stop_timer()

    def __draw_individual_lines(self, animal_img_data: pd.DataFrame, img: np.array):
        color = self.direction_colors[0]
        for cnt, (i, r) in enumerate(animal_img_data.iterrows()):
            if self.style_attr["Direction_color"] == "Random":
                color = random.sample(self.direction_colors[0], 1)[0]
            cv2.line(
                img,
                (int(r["Eye_x"]), int(r["Eye_y"])),
                (int(r["Animal_2_bodypart_x"]), int(r["Animal_2_bodypart_y"])),
                color,
                self.style_attr["Direction_thickness"],
            )
            if self.style_attr["Highlight_endpoints"]:
                cv2.circle(
                    img,
                    (int(r["Eye_x"]), int(r["Eye_y"])),
                    self.style_attr["Pose_circle_size"] + 2,
                    color,
                    self.style_attr["Pose_circle_size"],
                )
                cv2.circle(
                    img,
                    (int(r["Animal_2_bodypart_x"]), int(r["Animal_2_bodypart_y"])),
                    self.style_attr["Pose_circle_size"] + 1,
                    color,
                    self.style_attr["Pose_circle_size"],
                )

        return img

    def __create_video(self):
        img_cnt = 0
        video_timer = SimbaTimer(start=True)
        video_timer.start_timer()
        color = self.direction_colors[0]
        while self.cap.isOpened():
            ret, img = self.cap.read()
            try:
                if ret:
                    if self.style_attr["Show_pose"]:
                        bp_data = self.data_df.iloc[img_cnt]
                        for cnt, (animal_name, animal_bps) in enumerate(
                            self.animal_bp_dict.items()
                        ):
                            for bp in zip(animal_bps["X_bps"], animal_bps["Y_bps"]):
                                x_bp, y_bp = bp_data[bp[0]], bp_data[bp[1]]
                                cv2.circle(
                                    img,
                                    (int(x_bp), int(y_bp)),
                                    self.style_attr["Pose_circle_size"],
                                    self.animal_bp_dict[animal_name]["colors"][cnt],
                                    self.style_attr["Direction_thickness"],
                                )

                    if img_cnt in list(self.video_data["Frame_#"].unique()):
                        img_data = self.video_data[
                            self.video_data["Frame_#"] == img_cnt
                        ]
                        unique_animals = img_data["Animal_1"].unique()
                        for animal in unique_animals:
                            animal_img_data = img_data[
                                img_data["Animal_1"] == animal
                            ].reset_index(drop=True)
                            if self.style_attr["Polyfill"]:
                                convex_hull_arr = animal_img_data.loc[
                                    0, ["Eye_x", "Eye_y"]
                                ].values.reshape(-1, 2)
                                for animal_2 in animal_img_data["Animal_2"].unique():
                                    convex_hull_arr = np.vstack(
                                        (
                                            convex_hull_arr,
                                            animal_img_data[
                                                [
                                                    "Animal_2_bodypart_x",
                                                    "Animal_2_bodypart_y",
                                                ]
                                            ][
                                                animal_img_data["Animal_2"] == animal_2
                                            ].values,
                                        )
                                    ).astype("int")
                                    convex_hull_arr = np.unique(convex_hull_arr, axis=0)
                                    if convex_hull_arr.shape[0] >= 3:
                                        if (
                                            self.style_attr["Direction_color"]
                                            == "Random"
                                        ):
                                            color = random.sample(
                                                self.direction_colors[0], 1
                                            )[0]
                                        cv2.fillPoly(img, [convex_hull_arr], color)
                                    else:
                                        img = self.__draw_individual_lines(
                                            animal_img_data=animal_img_data, img=img
                                        )

                            else:
                                img = self.__draw_individual_lines(
                                    animal_img_data=animal_img_data, img=img
                                )

                    img_cnt += 1
                    self.writer.write(np.uint8(img))
                    print(
                        "Frame: {} / {}. Video: {}".format(
                            str(img_cnt),
                            str(self.video_meta_data["frame_count"]),
                            self.video_name,
                        )
                    )
                else:
                    self.cap.release()
                    self.writer.release()
                    break

            except IndexError:
                self.cap.release()
                self.writer.release()

        video_timer.stop_timer()
        stdout_success(
            msg=f"Directionality video {self.video_name} saved in {self.directing_animals_video_output_path} directory",
            elapsed_time=video_timer.elapsed_time_str,
        )


# style_attr = {'Show_pose': True, 'Pose_circle_size': 3, "Direction_color": 'Random', 'Direction_thickness': 4, 'Highlight_endpoints': True, 'Polyfill': True}
# test = DirectingOtherAnimalsVisualizer(config_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini',
#                                        data_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/csv/outlier_corrected_movement_location/Testing_Video_3.csv',
#                                        style_attr=style_attr)
#
# test.run()


# style_attr = {'Show_pose': True, 'Pose_circle_size': 3, "Direction_color": 'Random', 'Direction_thickness': 4, 'Highlight_endpoints': True, 'Polyfill': True}
# test = DirectingOtherAnimalsVisualizer(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                        data_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv',
#                                        style_attr=style_attr)
#
# test.run()



__author__ = "Simon Nilsson"

import os
from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, FolderSelect
from simba.utils.enums import Keys, Links
from simba.utils.errors import DuplicationError, NotDirectoryError
from simba.video_processors.batch_process_menus import BatchProcessFrame


class BatchPreProcessPopUp(PopUpMixin):
    def __init__(self):
        super().__init__(title="BATCH PROCESS VIDEO", size=(600, 400))
        selections_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SELECTIONS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.BATCH_PREPROCESS.value,
        )
        self.input_folder_select = FolderSelect(
            selections_frm,
            "INPUT VIDEO DIRECTORY:",
            title="Select Folder with Input Videos",
            lblwidth=20,
        )
        self.output_folder_select = FolderSelect(
            selections_frm,
            "OUTPUT VIDEO DIRECTORY:",
            title="Select Folder for Output videos",
            lblwidth=20,
        )
        confirm_btn = Button(
            selections_frm, text="CONFIRM", fg="blue", command=lambda: self.run()
        )
        selections_frm.grid(row=0, column=0, sticky=NW)
        self.input_folder_select.grid(row=0, column=0, sticky=NW)
        self.output_folder_select.grid(row=1, column=0, sticky=NW)
        confirm_btn.grid(row=2, column=0, sticky=NW)

    def run(self):
        if not os.path.isdir(self.input_folder_select.folder_path):
            raise NotDirectoryError(
                msg=f"INPUT folder dir ({self.input_folder_select.folder_path}) is not a valid directory.",
                source=self.__class__.__name__,
            )
        if not os.path.isdir(self.output_folder_select.folder_path):
            raise NotDirectoryError(
                msg=f"OUTPUT folder dir ({self.output_folder_select.folder_path}) is not a valid directory.",
                source=self.__class__.__name__,
            )
        if (
            self.output_folder_select.folder_path
            == self.input_folder_select.folder_path
        ):
            raise DuplicationError(
                msg="The INPUT directory and OUTPUT directory CANNOT be the same folder",
                source=self.__class__.__name__,
            )
        else:
            batch_preprocessor = BatchProcessFrame(
                input_dir=self.input_folder_select.folder_path,
                output_dir=self.output_folder_select.folder_path,
            )
            batch_preprocessor.create_main_window()
            batch_preprocessor.create_video_table_headings()
            batch_preprocessor.create_video_rows()
            batch_preprocessor.create_execute_btn()
            batch_preprocessor.main_frm.mainloop()



__author__ = "Simon Nilsson"

import os
from subprocess import PIPE, Popen
from tkinter import *
from tkinter import filedialog
from typing import Optional, Union

import cv2
import pandas as pd
from PIL import Image, ImageTk
from tabulate import tabulate

import simba
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_file_exist_and_readable, check_int
from simba.utils.enums import Formats, Options, TagNames
from simba.utils.errors import AdvancedLabellingError, FrameRangeError
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (get_all_clf_names, get_fn_ext,
                                    get_video_meta_data, read_config_entry,
                                    read_df, write_df)


class AdvancedLabellingInterface(ConfigReader):
    """
    Launch advanced labelling (annotation) interface in SimBA.

    .. note::
       `Advanced annotation tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/advanced_labelling.md>`__.

    .. image:: _static/img/adv_annotator.png
       :width: 800
       :align: center

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str file_path: Path to video that is to be annotated
    :parameter bool continuing: If True, user is continuing the annotations of a video with started but incomplete annotations.

    Examples
    ----------
    >>> select_labelling_video_advanced(config_path='MyProjectConfig', file_path='MyVideoFilePath', continuing=True)
    """

    def __init__(self, config_path: str, file_path: str, continuing: bool):
        ConfigReader.__init__(self, config_path=config_path)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.padding, self.file_path = 5, file_path
        self.frm_no, self.video_path = 0, file_path
        self.continuing = continuing
        self.play_video_script_path = os.path.join(
            os.path.dirname(simba.__file__), "labelling", "play_annotation_video.py"
        )
        _, self.video_name, _ = get_fn_ext(filepath=file_path)
        self.features_extracted_file_path = os.path.join(
            self.features_dir, self.video_name + "." + self.file_type
        )
        self.targets_inserted_file_path = os.path.join(
            self.targets_folder, self.video_name + "." + self.file_type
        )
        self.machine_results_file_path = os.path.join(
            self.machine_results_dir, self.video_name + "." + self.file_type
        )
        self.cap = cv2.VideoCapture(self.video_path)
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        self.frame_lst = list(range(0, self.video_meta_data["frame_count"]))
        self.max_frm_no = max(self.frame_lst)
        self.target_lst = get_all_clf_names(config=self.config, target_cnt=self.clf_cnt)
        self.max_frm_size = 1080, 650
        self.main_window = Toplevel()
        if continuing:
            check_file_exist_and_readable(file_path=self.targets_inserted_file_path)
            check_file_exist_and_readable(file_path=self.features_extracted_file_path)
            if self.file_type == Formats.CSV.value:
                self.data_df = (
                    pd.read_csv(self.targets_inserted_file_path)
                    .set_index("Unnamed: 0")[self.target_lst]
                    .astype(int)
                )
                self.data_df.index.name = None
            if self.file_type == Formats.PARQUET.value:
                self.data_df = pd.read_parquet(self.targets_inserted_file_path)[
                    self.target_lst
                ]
            self.data_df_features = read_df(
                self.features_extracted_file_path, self.file_type
            )
            self.data_df = self.data_df_features.join(self.data_df)
            self.main_window.title(
                "SIMBA ANNOTATION INTERFACE (CONTINUING ANNOTATIONS)"
            )
            try:
                self.frm_no = read_config_entry(
                    self.config,
                    "Last annotated frames",
                    self.video_name.lower(),
                    data_type="int",
                )
            except ValueError:
                pass
        else:
            check_file_exist_and_readable(file_path=self.features_extracted_file_path)
            self.data_df = read_df(self.features_extracted_file_path, self.file_type)
            for target in self.target_lst:
                self.data_df[target] = None
        self.main_window.title(
            "SIMBA ANNOTATION INTERFACE (ADVANCED ANNOTATION) - VIDEO {}".format(
                self.video_name
            )
        )
        self.data_df_targets = self.data_df[self.target_lst]
        self.folder = Frame(self.main_window)
        self.buttons_frm = Frame(self.main_window, bd=2, width=700, height=300)
        self.current_frm_n = IntVar(self.main_window, value=self.frm_no)
        self.change_frm_box = Entry(
            self.buttons_frm, width=7, textvariable=self.current_frm_n
        )
        self.frame_number_lbl = Label(self.buttons_frm, text="Frame number")
        self.forward_btn = Button(
            self.buttons_frm,
            text=">",
            command=lambda: self.advance_frame(
                new_frm_number=int(self.current_frm_n.get() + 1)
            ),
        )
        self.backward_btn = Button(
            self.buttons_frm,
            text="<",
            command=lambda: self.advance_frame(
                new_frm_number=int(self.current_frm_n.get() - 1)
            ),
        )
        self.forward_max_btn = Button(
            self.buttons_frm,
            text=">>",
            command=lambda: self.advance_frame(len(self.frame_lst) - 1),
        )
        self.next_absent_forward_btn = Button(
            self.buttons_frm,
            text="NEXT ABSENT",
            fg="red",
            command=lambda: self.find_last_next_annotation(
                forwards=True, present=False
            ),
        )
        self.next_present_forward_btn = Button(
            self.buttons_frm,
            text="NEXT PRESENT",
            fg="blue",
            command=lambda: self.find_last_next_annotation(forwards=True, present=True),
        )
        self.last_absent_forward_btn = Button(
            self.buttons_frm,
            text="PREVIOUS ABSENT",
            fg="red",
            command=lambda: self.find_last_next_annotation(
                forwards=False, present=False
            ),
        )
        self.last_present_forward_btn = Button(
            self.buttons_frm,
            text="PREVIOUS PRESENT",
            fg="blue",
            command=lambda: self.find_last_next_annotation(
                forwards=False, present=True
            ),
        )

        self.backward_max_btn = Button(
            self.buttons_frm, text="<<", command=lambda: self.advance_frame(0)
        )
        self.select_frm_btn = Button(
            self.buttons_frm,
            text="Jump to selected frame",
            command=lambda: self.advance_frame(
                new_frm_number=int(self.change_frm_box.get())
            ),
        )
        self.jump_frame = Frame(self.main_window)
        self.jump = Label(self.jump_frame, text="Jump Size:")
        self.jump_size = Scale(
            self.jump_frame, from_=0, to=100, orient=HORIZONTAL, length=200
        )
        self.jump_size.set(0)
        self.jump_back = Button(
            self.jump_frame,
            text="<<",
            command=lambda: self.advance_frame(
                int(self.change_frm_box.get()) - self.jump_size.get()
            ),
        )
        self.jump_forward = Button(
            self.jump_frame,
            text=">>",
            command=lambda: self.advance_frame(
                int(self.change_frm_box.get()) + self.jump_size.get()
            ),
        )

        self.folder.grid(row=0, column=1, sticky=N)
        self.buttons_frm.grid(row=1, column=0)
        self.change_frm_box.grid(row=0, column=1)
        self.last_present_forward_btn.grid(row=1, column=0, sticky=W, padx=self.padding)
        self.last_absent_forward_btn.grid(row=1, column=1, sticky=W, padx=self.padding)
        self.backward_max_btn.grid(row=1, column=2, sticky=W, padx=self.padding)
        self.backward_btn.grid(row=1, column=3, sticky=W, padx=self.padding)
        self.change_frm_box.grid(row=1, column=4)
        self.forward_btn.grid(row=1, column=5, sticky=E, padx=self.padding)
        self.forward_max_btn.grid(row=1, column=6, sticky=W, padx=self.padding)
        self.next_absent_forward_btn.grid(row=1, column=7, sticky=W, padx=self.padding)
        self.next_present_forward_btn.grid(row=1, column=8, sticky=W, padx=self.padding)

        self.select_frm_btn.grid(row=2, column=4, sticky=N)
        self.jump_frame.grid(row=2, column=0)
        self.jump.grid(row=0, column=0, sticky=W)
        self.jump_size.grid(row=0, column=1, sticky=W)
        self.jump_back.grid(row=0, column=2, sticky=E)
        self.jump_forward.grid(row=0, column=3, sticky=W)

        self.check_frame = Frame(self.main_window, bd=2, width=300, height=500)
        self.check_frame.grid(row=0, column=1)
        self.check_behavior_lbl = Label(self.check_frame, text="Check Behavior:")
        self.check_behavior_lbl.config(font=("Calibri", 16))
        self.check_behavior_lbl.grid(row=0, column=0, sticky=N)
        self.behavior_absent_lbl = Label(self.check_frame, text="ABSENT")
        self.behavior_present_lbl = Label(self.check_frame, text="PRESENT")
        self.behavior_present_lbl.grid(row=1, column=1, sticky=N)
        self.behavior_absent_lbl.grid(row=1, column=2, sticky=N)

        self.check_present_vars, self.check_present_checkbox = {}, {}
        self.check_absent_vars, self.check_absent_checkbox = {}, {}
        for target_cnt, target in enumerate(self.target_lst):
            self.check_present_vars[target], self.check_absent_vars[target] = (
                IntVar(),
                IntVar(),
            )
            self.behavior_name_lbl = Label(self.check_frame, text=target)
            self.check_present_checkbox[target] = Checkbutton(
                self.check_frame,
                variable=self.check_present_vars[target],
                command=lambda: self.save_behavior_in_frm(selection="present"),
            )
            self.check_absent_checkbox[target] = Checkbutton(
                self.check_frame,
                variable=self.check_absent_vars[target],
                command=lambda: self.save_behavior_in_frm(selection="absent"),
            )
            self.behavior_name_lbl.grid(row=target_cnt + 2, column=0, sticky=W)
            self.check_present_checkbox[target].grid(
                row=target_cnt + 2, column=1, sticky=W
            )
            self.check_absent_checkbox[target].grid(
                row=target_cnt + 2, column=2, sticky=W
            )
            if self.data_df_targets[target].iloc[self.current_frm_n.get()] == 1:
                self.check_present_vars[target].set(value=1)
            elif self.data_df_targets[target].iloc[self.current_frm_n.get()] == 0:
                self.check_absent_vars[target].set(value=1)

        self.range_on = IntVar(value=0)
        self.range_frames = Frame(self.main_window)
        self.range_frames.grid(row=1, column=1, sticky=S)
        self.select_range = Checkbutton(
            self.range_frames, text="Frame range", variable=self.range_on
        )
        self.select_range.grid(row=0, column=0, sticky=W)
        self.first_frame = Entry(self.range_frames, width=7)
        self.first_frame.grid(row=0, column=1, sticky=E)
        self.to_label = Label(self.range_frames, text=" to ")
        self.to_label.grid(row=0, column=2, sticky=E)
        self.last_frame = Entry(self.range_frames, width=7)
        self.last_frame.grid(row=0, column=3, sticky=E)

        save = Button(
            self.main_window,
            text="Save Range",
            command=lambda: self.save_behavior_in_range(),
        )
        save.grid(row=2, column=1, sticky=N)

        self.generate = Button(
            self.main_window,
            text="Save Annotations",
            command=lambda: self.save_results(),
            fg="blue",
        )
        self.generate.config(font=("Calibri", 16))
        self.generate.grid(row=10, column=1, sticky=N)

        self.video_player_frm = Frame(self.main_window, width=100, height=100)
        self.video_player_frm.grid(row=0, column=2, sticky=N)
        self.play_video_btn = Button(
            self.video_player_frm, text="Open Video", command=self.play_video
        )
        self.play_video_btn.grid(sticky=N, pady=10)
        self.video_key_lbls = Label(
            self.video_player_frm,
            text="\n\n  Keyboard shortcuts for video navigation: \n p = Pause/Play"
            "\n\n After pressing pause:"
            "\n o = +2 frames \n e = +10 frames \n w = +1 second"
            "\n\n t = -2 frames \n s = -10 frames \n x = -1 second"
            "\n\n q = Close video window \n\n",
        )
        self.video_key_lbls.grid(sticky=W)
        self.update_img_from_video = Button(
            self.video_player_frm,
            text="Show current video frame",
            command=self.update_frame_from_video,
        )
        self.update_img_from_video.grid(sticky=N)
        self.bind_shortcut_keys()
        self.key_presses_lbl = Label(
            self.video_player_frm,
            text="\n\n Keyboard shortcuts for frame navigation: \n Right Arrow = +1 frame"
            "\n Left Arrow = -1 frame"
            "\n Ctrl + s = Save annotations file"
            "\n Ctrl + a = +1 frame and keep choices"
            "\n Ctrl + p = Show annotation statistics"
            "\n Ctrl + l = Last frame"
            "\n Ctrl + o = First frame",
        )
        self.key_presses_lbl.grid(sticky=S)
        self.read_frm(frm_number=0)
        self.main_window.mainloop()

    def bind_shortcut_keys(self):
        self.main_window.bind("<Control-s>", lambda x: self.save_results())
        self.main_window.bind(
            "<Control-a>",
            lambda x: self.advance_frame(
                new_frm_number=int(self.current_frm_n.get() + 1),
                keep_prior_img_cb_status=True,
            ),
        )
        self.main_window.bind(
            "<Control-p>", lambda x: self.__print_annotation_statistics()
        )
        self.main_window.bind(
            "<Right>",
            lambda x: self.advance_frame(
                new_frm_number=int(self.current_frm_n.get() + 1)
            ),
        )
        self.main_window.bind(
            "<Left>",
            lambda x: self.advance_frame(
                new_frm_number=int(self.current_frm_n.get() - 1)
            ),
        )
        self.main_window.bind(
            "<Control-l>", lambda x: self.advance_frame(new_frm_number=self.max_frm_no)
        )
        self.main_window.bind("<Control-o>", lambda x: self.advance_frame(0))

    def find_last_next_annotation(self, forwards: bool, present: bool):
        if forwards:
            sliced = self.data_df_targets.loc[self.current_frm_n.get() + 1 :,].sum(
                axis=1
            )
            if present:
                frms = list(sliced.index[sliced >= 1])
            else:
                frms = list(sliced.index[sliced == 0])
            if len(frms) > 0:
                frm = frms[0]
            else:
                raise FrameRangeError(
                    "No forwards frames with annotation detected",
                    source=self.__class__.__name__,
                )
        else:
            sliced = self.data_df_targets.loc[0 : self.current_frm_n.get() - 1,].sum(
                axis=1
            )
            if present:
                frms = list(sliced.index[sliced >= 1])
            else:
                frms = list(sliced.index[sliced == 0])
            if len(frms) > 0:
                frm = frms[-1]
            else:
                raise FrameRangeError(
                    "No backwards frames with annotation detected",
                    source=self.__class__.__name__,
                )

        self.advance_frame(new_frm_number=frm, keep_prior_img_cb_status=False)

    def __print_annotation_statistics(self):
        table_view = [
            ["Video name", self.video_name],
            ["Video frames", self.video_meta_data["frame_count"]],
        ]
        for target in self.target_lst:
            present = sum(
                self.data_df_targets[target][self.data_df_targets[target] == 1]
            )
            absent = sum(
                self.data_df_targets[target][self.data_df_targets[target] == 0]
            )
            table_view.append([target + " present labels", present])
            table_view.append([target + " absent labels", absent])
            table_view.append(
                [target + " % present", present / self.video_meta_data["frame_count"]]
            )
            table_view.append(
                [target + " % absent", absent / self.video_meta_data["frame_count"]]
            )
        headers = ["VARIABLE", "VALUE"]
        print(tabulate(table_view, headers, tablefmt="github"))

    def play_video(self):
        p = Popen(
            "python {}".format(self.play_video_script_path),
            stdin=PIPE,
            stdout=PIPE,
            shell=True,
        )
        main_project_dir = os.path.dirname(self.config_path)
        p.stdin.write(bytes(self.video_path, "utf-8"))
        p.stdin.close()
        temp_file = os.path.join(main_project_dir, "subprocess.txt")
        with open(temp_file, "w") as text_file:
            text_file.write(str(p.pid))

    def update_frame_from_video(self):
        f = open(
            os.path.join(os.path.dirname(self.config_path), "labelling_info.txt"), "r+"
        )
        os.fsync(f.fileno())
        vid_frame_no = int(f.readline())
        self.advance_frame(new_frm_number=vid_frame_no)
        f.close()

    def read_frm(self, frm_number: int):
        self.cap.set(1, frm_number)
        _, self.current_frm_npy = self.cap.read()
        self.current_frm_npy = cv2.cvtColor(self.current_frm_npy, cv2.COLOR_RGB2BGR)
        self.current_frm_pil = Image.fromarray(self.current_frm_npy)
        self.current_frm_pil.thumbnail(self.max_frm_size, Image.ANTIALIAS)
        self.current_frm_pil = ImageTk.PhotoImage(
            master=self.main_window, image=self.current_frm_pil
        )
        self.video_frame = Label(self.main_window, image=self.current_frm_pil)
        self.video_frame.image = self.current_frm_pil
        self.video_frame.grid(row=0, column=0)

    def advance_frame(
        self,
        new_frm_number: int,
        save_behavior_in_previous_frm: bool = True,
        keep_prior_img_cb_status: bool = False,
    ):
        self.check_integrity_of_multiple_classifiers()
        if new_frm_number > self.max_frm_no:
            print(
                "FRAME {} CANNOT BE SHOWN - YOU ARE VIEWING THE FINAL FRAME OF THE VIDEO (FRAME NUMBER {})".format(
                    str(new_frm_number), str(self.max_frm_no)
                )
            )
            self.current_frm_n = IntVar(value=self.max_frm_no)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, self.current_frm_n.get())
        elif new_frm_number < 0:
            print(
                "FRAME {} CANNOT BE SHOWN - YOU ARE VIEWING THE FIRST FRAME OF THE VIDEO (FRAME NUMBER {})".format(
                    str(new_frm_number), str(self.max_frm_no)
                )
            )
            self.current_frm_n = IntVar(value=0)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, self.current_frm_n.get())
        elif (new_frm_number != self.current_frm_n.get()) and (
            not keep_prior_img_cb_status
        ):
            if save_behavior_in_previous_frm:
                self.save_behavior_in_frm()
            self.create_print_statements()
            self.current_frm_n = IntVar(value=new_frm_number)
            for target in self.target_lst:
                new_frame_annotation = self.data_df_targets[target].loc[
                    int(self.current_frm_n.get())
                ]
                if new_frame_annotation == 0:
                    self.check_absent_vars[target].set(value=1)
                    self.check_present_vars[target].set(value=0)
                elif new_frame_annotation == 1:
                    self.check_present_vars[target].set(value=1)
                    self.check_absent_vars[target].set(value=0)
                else:
                    self.check_present_vars[target].set(value=0)
                    self.check_absent_vars[target].set(value=0)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, new_frm_number)
            self.read_frm(frm_number=int(new_frm_number))

        elif (new_frm_number != self.current_frm_n.get()) and (
            keep_prior_img_cb_status
        ):
            self.create_print_statements()
            if save_behavior_in_previous_frm:
                self.save_behavior_in_frm()
            self.current_frm_n = IntVar(value=new_frm_number)
            for target in self.target_lst:
                new_frame_annotation = self.data_df_targets[target].loc[
                    int(self.current_frm_n.get() - 1)
                ]
                if new_frame_annotation == 0:
                    self.check_absent_vars[target].set(value=1)
                elif new_frame_annotation == 1:
                    self.check_present_vars[target].set(value=1)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, new_frm_number)
            self.read_frm(frm_number=int(new_frm_number))

    def save_behavior_in_frm(self, selection: str = None):
        for target in self.target_lst:
            target_absent_choice, target_present_choice = (
                self.check_absent_vars[target].get(),
                self.check_present_vars[target].get(),
            )
            if (target_present_choice == 1) & (selection == "present"):
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = 1
                self.check_absent_vars[target].set(value=0)
            elif (target_absent_choice == 1) & (selection == "absent"):
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = 0
                self.check_present_vars[target].set(value=0)
            elif (target_absent_choice == 0) & (target_present_choice == 0):
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = None
            elif target_present_choice == 1:
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = 1
            elif target_absent_choice == 1:
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = 0

    def save_behavior_in_range(self):
        self.check_integrity_of_multiple_classifiers()
        start_frm, end_frm = int(self.first_frame.get()), int(self.last_frame.get())
        check_int("START FRAME", int(start_frm), max_value=self.max_frm_no, min_value=0)
        check_int("END FRAME", int(end_frm), max_value=self.max_frm_no, min_value=0)
        if not self.range_on.get():
            raise FrameRangeError(
                msg="TO SAVE RANGE OF FRAMES, TICK THE `Frame range` checkbox before clicking `Save Range`",
                source=self.__class__.__name__,
            )
        elif start_frm < 0:
            raise FrameRangeError(
                msg="FRAME RANGE ERROR: START FRAME {} IS LESS THAN ZERO AND CANNOT BE SHOWN".format(
                    str(start_frm)
                ),
                source=self.__class__.__name__,
            )
        elif end_frm > self.max_frm_no:
            raise FrameRangeError(
                msg="FRAME RANGE ERROR: END FRAME {} IS MORE THAN THE MAX VIDEO FRAME ({}) AND CANNOT BE SHOWN".format(
                    str(end_frm), str(self.max_frm_no)
                ),
                source=self.__class__.__name__,
            )
        elif start_frm == end_frm:
            raise FrameRangeError(
                msg="FRAME RANGE ERROR: START FRAME AND END FRAME IS SET TO THE SAME VALUE ({}) AND DOES NOT REPRESENT A RANGE".format(
                    str(end_frm)
                ),
                source=self.__class__.__name__,
            )
        elif start_frm > end_frm:
            raise FrameRangeError(
                msg="FRAME RANGE ERROR: START FRAME ({}) IS LARGER THAB THE END FRAME ({}). PLEASE SPECIFY A RANGE OF FRAMES WHERE THE START FRAME PRECEDE THE END FRAME".format(
                    str(start_frm), str(end_frm)
                ),
                source=self.__class__.__name__,
            )
        else:
            for frm_no in range(int(start_frm), int(end_frm) + 1):
                for target in self.target_lst:
                    target_absent_choice, target_present_choice = (
                        self.check_absent_vars[target].get(),
                        self.check_present_vars[target].get(),
                    )
                    if target_present_choice == 1:
                        self.data_df_targets[target].loc[frm_no] = 1
                    if target_absent_choice == 1:
                        self.data_df_targets[target].loc[frm_no] = 0
                    if (target_absent_choice == 0) & (target_present_choice == 0):
                        self.data_df_targets[target].loc[frm_no] = None
            self.read_frm(frm_number=int(end_frm))
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, end_frm)
            self.create_print_statements(
                frame_range=True, start_frame=start_frm, end_frame=end_frm
            )

    def save_results(self):
        self.save_df = read_df(self.features_extracted_file_path, self.file_type)
        self.save_df = pd.concat([self.save_df, self.data_df_targets], axis=1)
        self.save_df = self.save_df.dropna(subset=self.target_lst)
        try:
            if self.file_type == Formats.CSV.value:
                self.save_df.to_csv(self.targets_inserted_file_path)
            if self.file_type == Formats.PARQUET.value:
                self.save_df.to_parquet(self.targets_inserted_file_path)
        except Exception as e:
            print(e, "SIMBA ERROR: File for video {} could not be saved.")
            raise FileExistsError
        stdout_success(
            msg=f"SAVED: Annotation file for video {self.video_name} saved within the project_folder/csv/targets_inserted directory.",
            source=self.__class__.__name__,
        )
        if not self.config.has_section("Last annotated frames"):
            self.config.add_section("Last annotated frames")
        self.config.set(
            "Last annotated frames", str(self.video_name), str(self.current_frm_n.get())
        )
        with open(self.config_path, "w") as configfile:
            self.config.write(configfile)

    def check_integrity_of_multiple_classifiers(self):
        none_target_lst, labelled_target_lst = [], []
        for target in self.target_lst:
            target_absent_choice, target_present_choice = (
                self.check_absent_vars[target].get(),
                self.check_present_vars[target].get(),
            )
            if (target_absent_choice == 0) and (target_present_choice == 0):
                none_target_lst.append(target)
            else:
                labelled_target_lst.append(target)
        if (len(none_target_lst) > 0) & (len(labelled_target_lst) > 0):
            raise AdvancedLabellingError(
                frame=str(self.current_frm_n.get()),
                lbl_lst=labelled_target_lst,
                unlabel_lst=none_target_lst,
                source=self.__class__.__name__,
            )

    def create_print_statements(
        self, frame_range: bool = None, start_frame: int = None, end_frame: int = None
    ):
        print("USER FRAME SELECTION(S):")
        if not frame_range:
            for target in self.target_lst:
                target_absent_choice, target_present_choice = (
                    self.check_absent_vars[target].get(),
                    self.check_present_vars[target].get(),
                )
                if (target_absent_choice == 1) & (target_present_choice == 0):
                    print(
                        "{} ABSENT IN FRAME {}".format(target, self.current_frm_n.get())
                    )
                if (target_present_choice == 1) & (target_absent_choice == 0):
                    print(
                        "{} PRESENT IN FRAME {}".format(
                            target, self.current_frm_n.get()
                        )
                    )
                if (target_present_choice == 0) & (target_absent_choice == 0):
                    print(
                        "{} UN-ANNOTATED IN FRAME {}".format(
                            target, self.current_frm_n.get()
                        )
                    )

        if frame_range:
            for target in self.target_lst:
                target_absent_choice, target_present_choice = (
                    self.check_absent_vars[target].get(),
                    self.check_present_vars[target].get(),
                )
                if (target_present_choice == 1) & (target_absent_choice == 0):
                    print(
                        "{} PRESENT IN FRAMES {} to {}".format(
                            target, str(start_frame), str(end_frame)
                        )
                    )
                elif (target_absent_choice == 1) & (target_present_choice == 0):
                    print(
                        "{} ABSENT IN FRAMES {} to {}".format(
                            target, str(start_frame), str(end_frame)
                        )
                    )
                elif (target_absent_choice == 0) & (target_present_choice == 0):
                    print(
                        "{} UN-ANNOTATED IN FRAMES {} to {}".format(
                            target, str(start_frame), str(end_frame)
                        )
                    )


def select_labelling_video_advanced(
    config_path: Union[str, os.PathLike], continuing: Optional[bool] = False
):
    video_file_path = filedialog.askopenfilename(
        filetypes=[("Video files", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)]
    )
    check_file_exist_and_readable(video_file_path)
    video_meta = get_video_meta_data(video_file_path)
    _, video_name, _ = get_fn_ext(video_file_path)
    print(f"ANNOTATING VIDEO {video_name} \n VIDEO INFO: {video_meta}")
    _ = AdvancedLabellingInterface(
        config_path=config_path, file_path=video_file_path, continuing=continuing
    )


# test = select_labelling_video_advanced(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                           continuing=False)



__author__ = "Simon Nilsson"

import glob
import os

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_that_column_exist)
from simba.utils.printing import stdout_success
from simba.utils.read_write import (get_fn_ext, read_config_file, read_df,
                                    write_df)


class ImportEthovision(ConfigReader):
    """
    Append ETHOVISION human annotations onto featurized pose-estimation data.
    Results are saved within the project_folder/csv/targets_inserted directory of
    the SimBA project (as parquets' or CSVs).

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_dir: path to folder holding ETHOVISION data files is XLSX or XLS format

    .. note::
       `Third-party import GitHub tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot.md>`__.
       `Example of expected ETHOVISION file <https://github.com/sgoldenlab/simba/blob/master/misc/ethovision_example.xlsx>`__.

    Examples
    -----
    >>> ethovision_importer = ImportEthovision(config_path="MyConfigPath", data_dir="MyEthovisionFolderPath")
    >>> ethovision_importer.run()
    """

    def __init__(self, config_path: str, data_dir: str):
        super().__init__(config_path=config_path)

        print("Appending ETHOVISION annotations...")
        self.config = read_config_file(config_path)
        self.files_found = glob.glob(data_dir + "/*.xlsx") + glob.glob(
            data_dir + "/*.xls"
        )
        self.files_found = [x for x in self.files_found if "~$" not in x]
        check_if_filepath_list_is_empty(
            filepaths=self.files_found,
            error_msg="SIMBA ERROR: No ETHOVISION xlsx or xls files found in {}".format(
                str(data_dir)
            ),
        )

    def __read_files(self):
        for file_path in self.files_found:
            ethovision_df = pd.read_excel(file_path, sheet_name=None)
            manual_scoring_sheet_name = list(ethovision_df.keys())[-1]
            ethovision_df = pd.read_excel(
                file_path,
                sheet_name=manual_scoring_sheet_name,
                index_col=0,
                header=None,
            )
            try:
                video_path = ethovision_df.loc["Video file"].values[0]
            except KeyError:
                print(
                    'SIMBA ERROR: "Video file" row does not exist in the sheet named {} in file {}'.format(
                        manual_scoring_sheet_name, file_path
                    )
                )
                raise ValueError(
                    f'SIMBA ERROR: "Video file" does not exist in the sheet named {manual_scoring_sheet_name} in file {file_path}'
                )
            try:
                if np.isnan(video_path):
                    print(
                        'SIMBA ERROR: "Video file" row does not have a value in the sheet named {} in file {}'.format(
                            manual_scoring_sheet_name, file_path
                        )
                    )
                    raise ValueError(
                        f'SIMBA ERROR: "Video file" row does not have a value in the sheet named {manual_scoring_sheet_name} in file {file_path}'
                    )
            except:
                pass
            dir_name, self.video_name, ext = get_fn_ext(video_path)
            self.processed_videos.append(video_path)
            self.features_file_path = os.path.join(
                self.features_dir, self.video_name + "." + self.file_type
            )
            print("Processing annotations for video " + str(self.video_name) + "...")
            _, _, fps = self.read_video_info(video_name=str(self.video_name))
            header_lines_n = (
                int(ethovision_df.loc["Number of header lines:"].values[0]) - 2
            )
            ethovision_df = ethovision_df.iloc[header_lines_n:].reset_index(drop=True)
            ethovision_df.columns = list(ethovision_df.iloc[0])
            ethovision_df = ethovision_df.iloc[2:].reset_index(drop=True)
            self.clf_dict = {}
            check_that_column_exist(
                df=ethovision_df, column_name="Behavior", file_name=file_path
            )
            check_that_column_exist(
                df=ethovision_df, column_name="Recording time", file_name=file_path
            )
            check_that_column_exist(
                df=ethovision_df, column_name="Event", file_name=file_path
            )
            non_clf_behaviors = list(
                set(ethovision_df["Behavior"].unique()) - set(self.clf_names)
            )
            non_clf_behaviors = [x for x in non_clf_behaviors if x.lower() != "start"]
            if len(non_clf_behaviors) > 0:
                print(
                    f"SIMBA WARNING: The ETHOVISION annotation file for video {self.video_name} contains annotations for {str(len(non_clf_behaviors))} behaviors"
                    f" which is NOT defined in the SimBA project: {non_clf_behaviors} and will be SKIPPED."
                )
            for clf in self.clf_names:
                self.clf_dict[clf] = {}
                clf_data = ethovision_df[ethovision_df["Behavior"] == clf]
                if len(clf_data) == 0:
                    print(
                        f"SIMBA WARNING: ZERO ETHOVISION annotations detected for SimBA classifier named {clf} for video {self.video_name}. "
                        f"SimBA will label that the behavior as ABSENT in the entire {self.video_name} video."
                    )
                starts = list(
                    clf_data["Recording time"][clf_data["Event"] == "state start"]
                )
                ends = list(
                    clf_data["Recording time"][clf_data["Event"] == "state stop"]
                )
                self.clf_dict[clf]["start_frames"] = [int(x * fps) for x in starts]
                self.clf_dict[clf]["end_frames"] = [int(x * fps) for x in ends]
                frame_list = []
                for cnt, start in enumerate(self.clf_dict[clf]["start_frames"]):
                    frame_list.extend(
                        list(range(start, self.clf_dict[clf]["end_frames"][cnt]))
                    )
                self.clf_dict[clf]["frames"] = frame_list
            self.__insert_annotations()

    def __insert_annotations(self):
        self.features_df = read_df(self.features_file_path, self.file_type)
        for clf in self.clf_names:
            annotation_mismatch = list(
                set(self.clf_dict[clf]["frames"]) - set(self.features_df.index)
            )
            if len(annotation_mismatch) > 0:
                print(
                    f"SIMBA ETHOVISION WARNING: SimBA found ETHOVISION annotations for behavior {clf} in video "
                    f"{self.video_name} that are annotated to occur at times which is not present in the "
                    f"video data you imported into SIMBA. The video you imported to SimBA has {str(max(self.features_df.index))} frames. "
                    f"However, in ETHOVISION, you have annotated {clf} to happen at frame number {str(annotation_mismatch[0])}. "
                    f"These ambiguous annotations occur in {str(len(annotation_mismatch))} different frames for video {self.video_name} that SimBA will **remove** by default. "
                    f"Please make sure you imported the same video as you annotated in ETHOVISION into SimBA and the video is registered with the correct frame rate."
                )
            self.features_df[clf] = 0
            self.features_df[clf] = np.where(
                self.features_df.index.isin(self.clf_dict[clf]["frames"]), 1, 0
            )
        self.__save_data()

    def __save_data(self):
        save_file_name = os.path.join(
            self.targets_folder, self.video_name + "." + self.file_type
        )
        write_df(self.features_df, self.file_type, save_file_name)
        print("Added Ethovision annotations for video {} ... ".format(self.video_name))

    def run(self):
        self.processed_videos = []
        self.__read_files()
        self.timer.stop_timer()
        stdout_success(
            msg="All Ethovision annotations added. Files with annotation are located in the project_folder/csv/targets_inserted directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = ImportEthovision(config_path= r"/Users/simon/Desktop/envs/simba_dev/tests/test_data/import_tests/project_folder/project_config.ini", data_dir=r'/Users/simon/Desktop/envs/simba_dev/tests/test_data/import_tests/ethovision_data')
# test = ImportEthovision(config_path= r"/Users/simon/Desktop/envs/simba_dev/test/data/test_projects/two_c57/project_folder/project_config.ini", data_dir='/Users/simon/Desktop/envs/simba_dev/test/data/test_projects/two_c57/ethovision_annotations')
# test.run()



__author__ = "Simon Nilsson"

import glob
import os
from copy import deepcopy

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.errors import AnnotationFileNotFoundError
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df
from simba.utils.warnings import (
    ThirdPartyAnnotationsAdditionalClfWarning,
    ThirdPartyAnnotationsClfMissingWarning,
    ThirdPartyAnnotationsOutsidePoseEstimationDataWarning)


class BentoAppender(ConfigReader):
    """
    Append BENTO annotation to SimBA featurized datasets.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_dir: Path to folder containing BENTO data.

    .. note::
       `Example BENTO input file <https://github.com/sgoldenlab/simba/blob/master/misc/bento_example.annot>`_.

       'GitHub tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot.md>`_.

    Examples
    ----------
    >>> bento_dir = 'tests/test_data/bento_example'
    >>> config_path = 'tests/test_data/import_tests/project_folder/project_config.ini'
    >>> bento_appender = BentoAppender(config_path=config_path, data_dir=bento_dir)
    >>> bento_appender.run()

    References
    ----------

    .. [1] Segalin et al., eLife, https://doi.org/10.7554/eLife.63720
    """

    def __init__(self, config_path: str, data_dir: str):
        ConfigReader.__init__(self, config_path=config_path)
        self.bento_dir = data_dir
        self.feature_files = glob.glob(self.features_dir + "/*." + self.file_type)
        self.bento_files = glob.glob(self.bento_dir + "/*." + "annot")
        check_if_filepath_list_is_empty(
            filepaths=self.feature_files,
            error_msg="SIMBA ERROR: No feature files found in project_folder/csv/features_extracted. Extract Features BEFORE appending BENTO annotations",
        )
        check_if_filepath_list_is_empty(
            filepaths=self.bento_files,
            error_msg=f"SIMBA ERROR: No BENTO files with .annot extension found in {self.bento_dir}.",
        )
        self.saved_files = []

    def run(self):
        for file_cnt, file_path in enumerate(self.feature_files):
            _, self.video_name, ext = get_fn_ext(filepath=file_path)
            print(f"Appending BENTO annotation to video {self.video_name}...")
            _, _, fps = self.read_video_info(video_name=self.video_name)
            bento_path = os.path.join(self.bento_dir, self.video_name + ".annot")
            if bento_path not in self.bento_files:
                raise AnnotationFileNotFoundError(video_name=self.video_name)
            self.save_path = os.path.join(
                self.targets_folder, self.video_name + "." + self.file_type
            )
            feature_df = read_df(file_path=file_path, file_type=self.file_type)
            video_frm_length = len(feature_df)
            self.results_df = deepcopy(feature_df)
            annotation_df = pd.read_csv(
                bento_path, delim_whitespace=True, index_col=False, low_memory=False
            )
            start_idx = annotation_df.index[
                annotation_df["Bento"] == "Ch1----------"
            ].values[0]
            sliced_annot = annotation_df.iloc[start_idx + 1 :]
            annotated_behaviors = sliced_annot[sliced_annot["Bento"].str.contains(">")][
                "Bento"
            ].tolist()
            annotated_behavior_names = [x[1:] for x in annotated_behaviors]
            missing_annotation = set(self.clf_names) - set(annotated_behavior_names)
            missing_clf = list(set(annotated_behavior_names) - set(self.clf_names))
            annotation_intersection = [
                x for x in self.clf_names if x in annotated_behavior_names
            ]
            for missing_clf in missing_annotation:
                ThirdPartyAnnotationsClfMissingWarning(
                    video_name=self.video_name, clf_name=missing_clf
                )
                self.results_df[missing_clf] = 0
            if missing_clf:
                ThirdPartyAnnotationsAdditionalClfWarning(
                    video_name=self.video_name, clf_names=missing_clf
                )

            for clf_name in annotation_intersection:
                self.results_df[clf_name] = 0
                clf_start_idx = sliced_annot.index[
                    sliced_annot["Bento"] == f">{clf_name}"
                ].values[0]
                clf_df = sliced_annot.loc[clf_start_idx + 2 :, :]
                end_idx = (
                    clf_df.isnull()[clf_df.isnull().any(axis=1)].idxmax(axis=1).index
                )
                if end_idx.values:
                    end_idx = end_idx.values[0]
                else:
                    end_idx = max(clf_df.index + 1)
                clf_df = clf_df.loc[: end_idx - 1, :].reset_index(drop=True)
                clf_df.columns = ["start_time", "stop_time", "duration"]
                clf_df["start_frm"] = clf_df["start_time"].astype(float) * fps
                clf_df["end_frm"] = clf_df["stop_time"].astype(float) * fps
                clf_df["start_frm"] = clf_df["start_frm"].astype(int)
                clf_df["end_frm"] = clf_df["end_frm"].astype(int)
                annotations_idx = list(
                    clf_df.apply(
                        lambda x: list(
                            range(int(x["start_frm"]), int(x["end_frm"]) + 1)
                        ),
                        1,
                    )
                )
                annotations_idx = [i for s in annotations_idx for i in s]
                annotations_idx_outside_video = [
                    x for x in annotations_idx if x > video_frm_length
                ]
                valid_annotation_ids = [
                    x for x in annotations_idx if x <= video_frm_length
                ]
                if len(annotations_idx_outside_video):
                    ThirdPartyAnnotationsOutsidePoseEstimationDataWarning(
                        video_name=self.video_name,
                        clf_name=clf_name,
                        frm_cnt=len(feature_df),
                        first_error_frm=annotations_idx_outside_video[0],
                        ambiguous_cnt=len(annotations_idx_outside_video),
                    )
                if len(valid_annotation_ids) > 0:
                    print(
                        f"Appending {str(len(valid_annotation_ids))} {clf_name} frame annotations to video {self.video_name}..."
                    )
                    self.results_df.loc[valid_annotation_ids, clf_name] = 1
            self.__save()
        stdout_success(
            msg=f"Annotations for {str(len(self.saved_files))} video(s) and saved in project_folder/csv/targets_inserted directory."
        )

    def __save(self):
        write_df(df=self.results_df, file_type=self.file_type, save_path=self.save_path)
        self.saved_files.append(self.save_path)
        print(
            f"BENTO annotations appended to video {self.video_name} and saved in {self.save_path}"
        )


# test = BentoAppender(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/import_tests/project_folder/project_config.ini',
#                      bento_dir='/Users/simon/Desktop/envs/simba_dev/tests/test_data/bento_example')
# test.run()
